"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
import copy
from collections import OrderedDict
from typing import Dict, List, Tuple, Union

from nagini_contracts.contracts import (
    CONTRACT_FUNCS,
    CONTRACT_WRAPPER_FUNCS
)
from nagini_contracts.io_contracts import IO_CONTRACT_FUNCS
from nagini_contracts.obligations import OBLIGATION_CONTRACT_FUNCS
from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.constants import (
    BUILTIN_PREDICATES,
    BUILTINS,
    DICT_TYPE,
    END_LABEL,
    ERROR_NAME,
    GET_ARG_FUNC,
    GET_METHOD_FUNC,
    GET_OLD_FUNC,
    INT_TYPE,
    JOINABLE_FUNC,
    LIST_TYPE,
    METHOD_ID_DOMAIN,
    OBJECT_TYPE,
    PRIMITIVE_INT_TYPE,
    PRIMITIVES,
    RANGE_TYPE,
    RESULT_NAME,
    SET_TYPE,
    STRING_TYPE,
    THREAD_DOMAIN,
    THREAD_POST_PRED,
    THREAD_START_PRED,
    TUPLE_TYPE,
)
from nagini_translation.lib.errors import rules
from nagini_translation.lib.program_nodes import (
    chain_cond_exp,
    GenericType,
    MethodType,
    OptionalType,
    PythonClass,
    PythonIOOperation,
    PythonMethod,
    PythonModule,
    PythonType,
    PythonVar,
    UnionType,
    toposort_classes,
    chain_if_stmts,
    SilverType,
)
from nagini_translation.lib.typedefs import (
    Expr,
    FuncApp,
    Position,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.util import (
    get_body_indices,
    get_func_name,
    InvalidProgramException,
    OldExpressionCollector,
    OldExpressionTransformer,
    pprint,
    UnsupportedException,
)
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator


class CallTranslator(CommonTranslator):


    def _translate_isinstance(self, node: ast.Call,
                              ctx: Context) -> StmtsAndExpr:
        assert len(node.args) == 2
        target = self.get_target(node.args[1], ctx)
        assert isinstance(target, (PythonType, PythonVar))
        stmt, obj = self.translate_expr(node.args[0], ctx)
        pos = self.to_position(node, ctx)
        if isinstance(target, PythonType):
            check = self.type_check(obj, target, pos, ctx, inhale_exhale=False)
        else:
            check = self.type_factory.dynamic_type_check(obj, target.ref(), pos,
                                                         ctx)
        return stmt, check

    def _translate_type_func(self, node: ast.Call,
                             ctx: Context) -> StmtsAndExpr:
        assert len(node.args) == 1
        stmt, obj = self.translate_expr(node.args[0], ctx)
        pos = self.to_position(node, ctx)
        result = self.type_factory.typeof(obj, ctx)
        return stmt, result

    def _translate_cast_func(self, node: ast.Call,
                             ctx: Context) -> StmtsAndExpr:
        stmt, object_arg = self.translate_expr(node.args[1], ctx)
        cast_type = self.get_type(node, ctx)
        arg_pos = self.to_position(node.args[0], ctx)
        type_arg = self.type_factory.translate_type_literal(cast_type,
                                                            arg_pos, ctx)
        object_class = ctx.module.global_module.classes['object']
        result = self.get_function_call(object_class, '__cast__',
                                        [type_arg, object_arg], [None, None],
                                        node, ctx)
        return stmt, result

    def _translate_len(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        assert len(node.args) == 1
        stmt, target = self.translate_expr(node.args[0], ctx)
        arg_type = self.get_type(node.args[0], ctx).try_box()
        len_stmt, len_val = self.get_func_or_method_call(arg_type, '__len__', [target],
                                                         [None], node, ctx)
        return stmt + len_stmt, len_val

    def _translate_str(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        assert len(node.args) == 1
        stmt, target = self.translate_expr(node.args[0], ctx)
        arg_type = self.get_type(node.args[0], ctx)
        str_stmt, str_val = self.get_func_or_method_call(arg_type, '__str__', [target],
                                                         [None], node, ctx)
        return stmt + str_stmt, str_val


    def _translate_int(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        assert len(node.args) == 1
        stmt, target = self.translate_expr(node.args[0], ctx)
        arg_type = self.get_type(node.args[0], ctx)
        str_stmt, str_val = self.get_func_or_method_call(arg_type, '__int__', [target],
                                                         [None], node, ctx)
        return stmt + str_stmt, str_val

    def _translate_bool(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        assert len(node.args) == 1
        stmt, target = self.translate_expr(node.args[0], ctx)
        arg_type = self.get_type(node.args[0], ctx)
        bool_stmt, bool_val = self.get_func_or_method_call(arg_type, '__bool__', [target],
                                                           [None], node, ctx)
        return stmt + bool_stmt, bool_val

    def _translate_super(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        This will only be called if super() is called in some place that is not the
        receiver of a method call. Who knows what happens when you interact with that;
        we do not support it.
        """
        raise InvalidProgramException(node, 'invalid.super.call')

    def translate_adt_cons(self, cons: PythonClass, args: List[FuncApp],
                           pos: Position, ctx: Context) -> Expr:
        """
        Constructs ADTs via a sequence of constructor calls and
        boxing/unboxing calls.
        """
        info = self.no_info(ctx)
        adt_type = self.viper.DomainType(cons.adt_domain_name, {}, [])

        # If expected argument type is the ADT type (another constructor call),
        # unbox translated argument
        for index, (arg_type, translated_arg) in enumerate(zip(cons.fields.values(),
                                                               args)):
            if arg_type.type == cons.adt_def:
                unbox_func = self.viper.FuncApp(cons.fresh('unbox_' +
                                                cons.adt_domain_name),
                                                [translated_arg], pos,
                                                info, adt_type)
                args[index] = unbox_func
            else:
                if arg_type.type.name in PRIMITIVES:
                    v_type = self.translate_type(arg_type.type, ctx)
                    args[index] = self.to_type(translated_arg, v_type, ctx)

        # Translate constructor call
        cons_call = self.viper.DomainFuncApp(cons.fresh(cons.adt_prefix +
                                             cons.name), args, adt_type,
                                             pos, info, cons.adt_domain_name)

        # Box translated constructor
        box_func = self.viper.FuncApp(cons.fresh('box_' + cons.adt_domain_name),
                                      [cons_call], pos, info, self.viper.Ref)

        return box_func

    def _is_lock_subtype(self, cls: PythonClass) -> bool:
        if cls is None:
            return False
        if cls.name == 'Lock':
            return cls
        return self._is_lock_subtype(cls.superclass)

    def translate_constructor_call(self, target_class: PythonClass,
            node: ast.Call, args: List, arg_stmts: List,
            ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the constructor of target_class with args, where
        node is the call node and arg_stmts are statements related to argument
        evaluation.
        """
        assert all(args), "Some args are None: {}".format(args)
        pos = self.to_position(node, ctx)

        if target_class.is_adt:
            return arg_stmts, self.translate_adt_cons(target_class, args, pos, ctx)

        res_var = ctx.current_function.create_variable(target_class.name +
                                                       '_res',
                                                       target_class,
                                                       self.translator)
        result_type = self.get_type(node, ctx)
        info = self.no_info(ctx)

        # Temporarily bind the type variables of the constructed class to
        # the concrete type arguments.
        old_bound_type_vars = ctx.bound_type_vars
        ctx.bound_type_vars = old_bound_type_vars.copy()
        current_type = result_type
        while current_type:
            if isinstance(current_type, GenericType):
                vars_args = zip(current_type.python_class.type_vars.items(),
                                current_type.type_args)
                for (name, var), arg in vars_args:
                    literal = self.type_factory.translate_type_literal(arg, pos,
                                                                       ctx)
                    key = (var.target_type.name, name)
                    ctx.bound_type_vars[key] = literal
            current_type = current_type.superclass

        fields = list(target_class.all_fields)
        may_set_inhales = [self.viper.Inhale(self.get_may_set_predicate(res_var.ref(),
                                                                        f, ctx),
                                             pos, self.no_info(ctx))
                           for f in fields]

        ctx.bound_type_vars = old_bound_type_vars
        new = self.viper.NewStmt(res_var.ref(), [], self.no_position(ctx),
                                 self.no_info(ctx))

        result_has_type = self.type_factory.type_check(res_var.ref(), result_type, pos,
                                                       ctx, concrete=True)
        defined_check = []
        if target_class.module is not target_class.module.global_module:
            # Mark the current function as depending on the called class. If we're in
            # a global context, assert that the called class and its dependencies are
            # defined.
            func_node = node.func if isinstance(node, ast.Call) else node
            self._add_dependencies(func_node, target_class, ctx)
            if self.is_main_method(ctx):
                defined_check = self.assert_global_defined(target_class, ctx.module,
                                                           node.func, ctx)

        # Inhale the type information about the newly created object
        # so that it's already present when calling __init__.
        type_inhale = self.viper.Inhale(result_has_type, pos,
                                        self.no_info(ctx))
        args = [res_var.ref()] + args
        stmts = [new, type_inhale] + may_set_inhales

        if self._is_lock_subtype(target_class):
            # For locks, fold the invariant predicate before actually calling the
            # constructor (which requires the predicate in its precondition).
            lock_pos = self.to_position(node, ctx, rules=rules.LOCK_RELEASE_INVARIANT)
            lock_class = self._is_lock_subtype(target_class)
            locked_obj = self.get_function_call(lock_class, 'get_locked', [res_var.ref()],
                                                [None], node, ctx, lock_pos)
            arg_is_locked = self.viper.EqCmp(locked_obj, args[1], lock_pos, info)
            stmts.append(self.viper.Inhale(arg_is_locked, lock_pos, info))
            invariant_pred = lock_class.get_predicate('invariant')
            full_perm = self.viper.FullPerm(lock_pos, info)
            pred_acc = self.viper.PredicateAccess([res_var.ref()],
                                                  invariant_pred.sil_name, lock_pos, info)
            acc_pred = self.viper.PredicateAccessPredicate(pred_acc, full_perm,
                                                           lock_pos, info)
            stmts.append(self.viper.Fold(acc_pred, lock_pos, info))

        target = target_class.get_method('__init__')
        if target:
            target_class = target.cls
            targets = []
            if target.declared_exceptions:
                error_var = self.get_error_var(node, ctx)
                targets.append(error_var)
            method_name = target_class.get_method('__init__').sil_name
            init = self.create_method_call_node(
                ctx, method_name, args, targets, self.to_position(node, ctx),
                self.no_info(ctx), target_method=target, target_node=node)
            stmts.extend(init)
            if target.declared_exceptions:
                catchers = self.create_exception_catchers(error_var,
                    ctx.actual_function.try_blocks, node, ctx)
                stmts = stmts + catchers
        return arg_stmts + defined_check + stmts, res_var.ref()

    def _translate_list(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        contents = None
        stmts = []
        if node.args:
            assert len(node.args) == 1
            arg_stmt, arg_val = self.translate_expr(node.args[0], ctx)
            stmts.extend(arg_stmt)
            contents = arg_val
        list_class = ctx.module.global_module.classes[LIST_TYPE]
        res_var = ctx.current_function.create_variable('list',
                                                       list_class, self.translator)
        targets = [res_var.ref()]
        constr_call = self.get_method_call(list_class, '__init__', [],
                                           [], targets, node, ctx)
        stmts.extend(constr_call)
        # Inhale the type of the newly created set (including type arguments)
        set_type = self.get_type(node, ctx)
        if (node._parent and isinstance(node._parent, ast.Assign) and
                    len(node._parent.targets) == 1):
            set_type = self.get_type(node._parent.targets[0], ctx)
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        result_var = res_var.ref(node, ctx)
        stmts.append(self.viper.Inhale(self.type_check(result_var,
                                                       set_type, position, ctx),
                                       position, self.no_info(ctx)))
        if contents:
            sil_ref_seq = self.viper.SeqType(self.viper.Ref)
            ref_seq = SilverType(sil_ref_seq, ctx.module)
            havoc_var = ctx.actual_function.create_variable('havoc_seq', ref_seq,
                                                            self.translator)
            seq_field = self.viper.Field('list_acc', sil_ref_seq, position, info)
            content_field = self.viper.FieldAccess(result_var, seq_field, position, info)
            stmts.append(self.viper.FieldAssign(content_field, havoc_var.ref(), position,
                                                info))
            arg_type = self.get_type(node.args[0], ctx)
            arg_seq = self.get_sequence(arg_type, contents, None, node, ctx, position)
            res_seq = self.get_sequence(set_type, result_var, None, node, ctx, position)
            seq_equal = self.viper.EqCmp(arg_seq, res_seq, position, info)
            stmts.append(self.viper.Inhale(seq_equal, position, info))
        return stmts, result_var

    def _translate_set(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        contents = None
        stmts = []
        if node.args:
            assert len(node.args) == 1
            arg_stmt, arg_val = self.translate_expr(node.args[0], ctx)
            stmts.extend(arg_stmt)
            contents = arg_val
        set_class = ctx.module.global_module.classes[SET_TYPE]
        res_var = ctx.current_function.create_variable('set',
            set_class, self.translator)
        targets = [res_var.ref()]
        constr_call = self.get_method_call(set_class, '__init__', [],
                                           [], targets, node, ctx)
        stmts.extend(constr_call)
        # Inhale the type of the newly created set (including type arguments)
        set_type = self.get_type(node, ctx)
        if (node._parent and isinstance(node._parent, ast.Assign) and
                len(node._parent.targets) == 1):
            set_type = self.get_type(node._parent.targets[0], ctx)
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        result_var = res_var.ref(node, ctx)
        stmts.append(self.viper.Inhale(self.type_check(result_var,
                                                       set_type, position, ctx),
                                       position, self.no_info(ctx)))
        if contents:
            sil_ref_set = self.viper.SetType(self.viper.Ref)
            ref_set = SilverType(sil_ref_set, ctx.module)
            havoc_var = ctx.actual_function.create_variable('havoc_set', ref_set,
                                                            self.translator)
            set_field = self.viper.Field('set_acc', sil_ref_set, position, info)
            content_field = self.viper.FieldAccess(result_var, set_field, position, info)
            stmts.append(self.viper.FieldAssign(content_field, havoc_var.ref(), position,
                                                info))
            arg_type = self.get_type(node.args[0], ctx)
            quant_var_name = ctx.actual_function.get_fresh_name('item')
            quant_var = self.viper.LocalVar(quant_var_name, self.viper.Ref, position,
                                            info)
            quant_var_decl = self.viper.LocalVarDecl(quant_var_name, self.viper.Ref,
                                                     position, info)
            arg_contains = self.get_function_call(arg_type, '__contains__',
                                                  [contents, quant_var], [None, None],
                                                  node, ctx)
            res_contains = self.get_function_call(set_type, '__contains__',
                                                  [result_var, quant_var], [None, None],
                                                  node, ctx)
            contain_equal = self.viper.EqCmp(arg_contains, res_contains, position, info)
            trigger = self.viper.Trigger([res_contains], position, info)
            quantifier = self.viper.Forall([quant_var_decl], [trigger], contain_equal,
                                           position, info)
            stmts.append(self.viper.Inhale(quantifier, position, info))
        return stmts, result_var

    def _translate_range(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        if len(node.args) != 2:
            msg = 'range() is currently only supported with two args.'
            raise UnsupportedException(node, msg)
        range_class = ctx.module.global_module.classes[RANGE_TYPE]
        start_stmt, start = self.translate_expr(node.args[0], ctx,
                                                self.viper.Int)
        end_stmt, end = self.translate_expr(node.args[1], ctx, self.viper.Int)
        # Add unique integer to make new instance different from other ranges.
        args = [start, end, self.get_fresh_int_lit(ctx)]
        arg_types = [None, None, None]
        call = self.get_function_call(range_class, '__create__', args,
                                      arg_types, node, ctx)
        return start_stmt + end_stmt, call

    def _translate_enumerate(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translate a call to enumerate(iterable) to a creation of a new list-object,
        an inhale about its type, and an inhale defining the contents of the new
        list.
        """
        if len(node.args) != 1:
            msg = 'enumerate() is currently only supported with one args.'
            raise UnsupportedException(node, msg)
        pos = self.to_position(node, ctx, rules=rules.INHALE_TO_CALL)
        info = self.no_info(ctx)
        result_type = self.get_type(node, ctx)
        arg_type = self.get_type(node.args[0], ctx)
        arg_stmt, arg = self.translate_expr(node.args[0], ctx)
        arg_contents = self.get_sequence(arg_type, arg, None, node.args[0], ctx)
        new_list = ctx.actual_function.create_variable('enumerate_res', result_type,
                                                       self.translator)
        sil_ref_seq = self.viper.SeqType(self.viper.Ref)
        seq_field = self.viper.Field('list_acc', sil_ref_seq, pos, info)
        new_stmt = self.viper.NewStmt(new_list.ref(), [seq_field], pos, info)
        seq_ref = self.viper.FieldAccess(new_list.ref(), seq_field, pos, info)
        prim_int_type = ctx.module.global_module.classes[PRIMITIVE_INT_TYPE]
        int_type = ctx.module.global_module.classes[INT_TYPE]
        list_type_info = self.type_check(new_list.ref(), result_type, pos, ctx)
        list_len_info = self.viper.EqCmp(self.viper.SeqLength(seq_ref, pos, info),
                                         self.viper.SeqLength(arg_contents, pos, info),
                                         pos, info)
        type_inhale = self.viper.Inhale(self.viper.And(list_type_info, list_len_info,
                                                       pos, info),
                                        pos, info)
        i_var = ctx.actual_function.create_variable('i', prim_int_type, self.translator,
                                                    False)
        orig_seq_i = self.viper.SeqIndex(arg_contents, i_var.ref(), pos, info)
        content_type = result_type.type_args[0].type_args[1]
        content_has_type = self.type_check(orig_seq_i, content_type, pos, ctx, False)
        tuple_for_i = self.create_tuple([i_var.ref(), orig_seq_i],
                                        [int_type, content_type], node, ctx)
        new_list_i = self.viper.SeqIndex(seq_ref, i_var.ref(), pos, info)
        equal_content = self.viper.EqCmp(new_list_i, tuple_for_i, pos, info)
        i_positive = self.viper.GeCmp(i_var.ref(), self.viper.IntLit(0, pos, info), pos,
                                      info)
        i_lt_len = self.viper.LtCmp(i_var.ref(), self.viper.SeqLength(seq_ref, pos, info),
                                    pos, info)
        i_in_bounds = self.viper.And(i_positive, i_lt_len, pos, info)
        content_info = self.viper.And(content_has_type, equal_content, pos, info)
        body = self.viper.Implies(i_in_bounds, content_info, pos, info)
        trigger = self.viper.Trigger([new_list_i], pos, info)
        contents_info = self.viper.Forall([i_var.decl], [trigger], body, pos, info)
        contents_inhale = self.viper.Inhale(contents_info, pos, info)
        return arg_stmt + [new_stmt, type_inhale, contents_inhale], new_list.ref(node,
                                                                                 ctx)

    def _translate_builtin_func(self, node: ast.Call,
                                ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to a builtin function like len() or isinstance()
        """
        func_name = get_func_name(node)
        if func_name == 'isinstance':
            return self._translate_isinstance(node, ctx)
        elif func_name == 'super':
            return self._translate_super(node, ctx)
        elif func_name == 'len':
            return self._translate_len(node, ctx)
        elif func_name == 'str':
            return self._translate_str(node, ctx)
        elif func_name == 'int':
            return self._translate_int(node, ctx)
        elif func_name == 'bool':
            return self._translate_bool(node, ctx)
        elif func_name == 'set':
            return self._translate_set(node, ctx)
        elif func_name == 'list':
            return self._translate_list(node, ctx)
        elif func_name == 'range':
            return self._translate_range(node, ctx)
        elif func_name == 'enumerate':
            return self._translate_enumerate(node, ctx)
        elif func_name == 'type':
            return self._translate_type_func(node, ctx)
        elif func_name == 'cast':
            return self._translate_cast_func(node, ctx)
        else:
            raise UnsupportedException(node)

    def _translate_method_call(self, target: PythonMethod, args: List[Expr],
                               arg_stmts: List[Stmt],
                               position: 'silver.ast.Position', node: ast.AST,
                               ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to an impure method, and performs any additional steps around
        it that might be necessary (e.g., folding/unfolding invariant predicates if the
        call is a lock release/acquire).
        """
        stmts = arg_stmts
        if target.name in ('acquire', 'release'):
            if (target.cls and target.cls.name == 'Lock' and
                        target.cls.module.type_prefix == 'nagini_contracts.lock'):
                # Store receiver for later use (contents of args list get changed by
                # subsequent call).
                receiver = args[0]
        call_stmt, res = self._only_translate_method_call(target, args, position, node,
                                                          ctx)
        if target.name in ('acquire', 'release'):
            if (target.cls and target.cls.name == 'Lock' and
                        target.cls.module.type_prefix == 'nagini_contracts.lock'):
                # Automatically fold/unfold the invariant predicate.
                target_name = target.cls.get_predicate('invariant').sil_name
                pos = self.to_position(node, ctx, rules=rules.LOCK_RELEASE_INVARIANT)
                info = self.no_info(ctx)
                pa = self.viper.PredicateAccess([receiver], target_name, pos, info)
                full_perm = self.viper.FullPerm(pos, info)
                pap = self.viper.PredicateAccessPredicate(pa, full_perm, pos, info)
                if target.name == 'acquire':
                    stmts.extend(call_stmt)
                    unfold = self.viper.Unfold(pap, pos, info)
                    stmts.append(unfold)
                else:
                    fold = self.viper.Fold(pap, pos, info)
                    stmts.append(fold)
                    stmts.extend(call_stmt)
        else:
            stmts.extend(call_stmt)
        return stmts, res

    def _only_translate_method_call(self, target: PythonMethod, args: List[Expr],
                                    position: 'silver.ast.Position', node: ast.AST,
                                    ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to an impure method.
        """
        targets = []
        result_var = None
        if target.type is not None:
            result_var = ctx.current_function.create_variable(
                target.name + '_res', target.type, self.translator)
            targets.append(result_var.ref())
        if target.declared_exceptions:
            error_var = self.get_error_var(node, ctx)
            targets.append(error_var)
        defined_check = []
        if target.module is not target.module.global_module:
            # Mark the current function as depending on the called method. If we're in
            # a global context, assert that the called method and its dependencies are
            # defined.
            self._add_dependencies(node.func, target, ctx)
            if self.is_main_method(ctx) and not target.cls:
                defined_check = self.assert_global_defined(target, ctx.module, node.func,
                                                           ctx)
        call = self.create_method_call_node(
            ctx, target.sil_name, args, targets, position, self.no_info(ctx),
            target_method=target, target_node=node)
        if target.declared_exceptions:
            call = call + self.create_exception_catchers(error_var,
                ctx.actual_function.try_blocks, node, ctx)
        return (defined_check + call,
                result_var.ref() if result_var else None)

    def _add_dependencies(self, reference: ast.AST, target: PythonMethod,
                          ctx: Context) -> None:
        """
        Tracks that the current container (method or class) depends on the given target,
        referring to it via the given reference.
        """
        if ctx.current_function and not self.is_main_method(ctx):
            if ctx.current_function:
                current_container = ctx.current_function.call_deps
            else:
                current_container = ctx.current_class.definition_deps
            if (isinstance(target, PythonMethod) and target.cls and
                        target.method_type == MethodType.normal):
                for subclass in target.cls.all_subclasses:
                    current = subclass.get_method(target.name)
                    current_container.add((reference, current, ctx.module, subclass))
            else:
                current_container.add((reference, target, ctx.module))

    def _translate_function_call(self, target: PythonMethod, args: List[Expr],
                                 formal_args: List[Expr], arg_stmts: List[Stmt],
                                 position: 'silver.ast.Position', node: ast.AST,
                                 ctx: Context) -> StmtsAndExpr:
        """Translates a call to a pure method."""
        type = self.translate_type(target.type, ctx)
        call = self.viper.FuncApp(target.sil_name, args, position,
                                  self.no_info(ctx), type, formal_args)
        if target.module is not target.module.global_module:
            # Mark the current function as depending on the called function. If we're in
            # a global context, wrap the result into a check that the called function and its
            # dependencies are defined.
            self._add_dependencies(node.func, target, ctx)
            if not target.cls and self.is_main_method(ctx):
                call = self.wrap_global_defined_check(call, target, ctx.module, node.func, ctx)
        return arg_stmts, call

    def _get_call_target(self, node: ast.Call,
                         ctx: Context) -> Union[PythonClass, PythonMethod]:
        """
        Returns the target of the given call; for constructor calls, the class
        whose constructor is called, for everything else the method.
        """
        target = self.get_target(node.func, ctx)
        if target:
            return target
        name = get_func_name(node)
        if name in ctx.module.classes:
            # Constructor call
            return ctx.module.classes[name]
        if isinstance(node.func, ast.Attribute):
            if isinstance(node.func.value, ast.Name):
                if node.func.value.id in ctx.module.classes:
                    # Statically bound call
                    target_class = ctx.module.classes[node.func.value.id]
                    return target_class.get_func_or_method(node.func.attr)
            if isinstance(node.func.value, ast.Call):
                if get_func_name(node.func.value) == 'super':
                    # Super call
                    target_class = self.get_target(node.func.value, ctx)
                    return target_class.get_func_or_method(node.func.attr)
            # Method called on an object
            receiver_class = self.get_type(node.func.value, ctx)
            # When receiver's type is union, a method call have multiple
            # targets, therefore None is returned in such cases
            if isinstance(receiver_class, UnionType):
                return None
            target = receiver_class.get_predicate(node.func.attr)
            if not target:
                target = receiver_class.get_func_or_method(node.func.attr)
            return target
        else:
            # Global function/method called
            receiver_class = None
            target = ctx.module.predicates.get(name)
            if not target:
                target = ctx.module.get_func_or_method(name)
            return target

    def _has_implicit_receiver_arg(self, node: ast.Call, ctx: Context) -> bool:
        """
        Checks if the given call node will have to have a receiver added to the
        arguments in the Silver encoding.
        """
        # Get target
        called_func = self.get_target(node.func, ctx)
        if isinstance(called_func, PythonClass):
            # constructor
            return True
        # If normal
        assert isinstance(called_func, PythonMethod)
        if (isinstance(node.func, ast.Attribute) and
                get_func_name(node.func.value) == 'super'):
            return True
        if called_func.method_type == MethodType.normal:
            if isinstance(node.func, ast.Attribute):
                called_name = get_func_name(node.func.value)
                if called_name == 'Result':
                    return True
                rec_target = self.get_target(node.func.value, ctx)
                if isinstance(rec_target, PythonModule):
                    return False
                elif (isinstance(rec_target, PythonClass) and
                      not isinstance(node.func.value, ast.Call) and
                      not isinstance(node.func.value, ast.Str)):
                    return False
                else:
                    return True
            else:
                return False
        elif called_func.method_type == MethodType.class_method:
            return True
        else:
            return False

    def _translate_call_args(self, node: ast.Call,
                             ctx: Context) -> Tuple[List[Stmt], List[Expr],
                                              List[PythonType]]:
        target = self._get_call_target(node, ctx)
        if isinstance(target, PythonClass):
            constr = target.get_method('__init__')
            target = constr
        return self.translate_args(target, node.args, node.keywords, node, ctx)

    def translate_args(self, target: PythonMethod, arg_nodes: List,
                       keywords: List, node: ast.AST, ctx: Context,
                       implicit_receiver=None) -> Tuple[List[Stmt], List[Expr],
                                                        List[PythonType]]:
        """
        Returns the args and types of the given call. Named args are put into
        the correct position; for *args and **kwargs, tuples and dicts are
        created and all arguments not bound to any other parameter are put in
        there.
        """
        arg_stmts = []

        unpacked_args = []
        unpacked_arg_types = []

        for arg in arg_nodes:
            if isinstance(arg, ast.Starred):
                # If it's a starred expression, unpack all values separately
                # into unpacked_args.
                arg_stmt, arg_expr = self.translate_expr(arg.value, ctx)
                arg_type = self.get_type(arg.value, ctx)
                arg_stmts += arg_stmt

                if (isinstance(arg_type, GenericType) and
                            arg_type.name == TUPLE_TYPE):
                    if not arg_type.exact_length:
                        raise UnsupportedException(arg, 'Starred expression '
                                                        'with statically '
                                                        'unknown length.')
                    nargs = len(arg_type.type_args)
                    for i, type_arg in enumerate(arg_type.type_args):
                        index = self.viper.IntLit(i, self.no_position(ctx),
                                                  self.no_info(ctx))
                        item = self.get_function_call(arg_type, '__getitem__',
                                                      [arg_expr, index],
                                                      [None, None], arg, ctx)
                        unpacked_args.append(item)
                        unpacked_arg_types.append(type_arg)
                else:
                    raise UnsupportedException(arg, 'Starred expression which '
                                                    'is not a tuple.')
            else:
                arg_stmt, arg_expr = self.translate_expr(arg, ctx)
                arg_type = self.get_type(arg, ctx)
                arg_stmts += arg_stmt
                unpacked_args.append(arg_expr)
                unpacked_arg_types.append(arg_type)

        if not target:
            return arg_stmts, unpacked_args, unpacked_arg_types

        if implicit_receiver is None:
            implicit_receiver = self._has_implicit_receiver_arg(node, ctx)

        if target.interface:
            # For interface functions defined directly in silver, missing arguments
            # are just set to null.
            if keywords:
                raise UnsupportedException(node, desc='Keyword arguments in call to '
                                                      'builtin function.')
            diff = target.nargs - len(unpacked_args)
            if diff < 0:
                raise UnsupportedException(node, 'Unsupported version of builtin '
                                                 'function.')
            if diff > 0:
                null = self.viper.NullLit(self.no_position(ctx), self.no_info(ctx))
                unpacked_args += [null] * diff
                unpacked_arg_types += [None] * diff
            return arg_stmts, unpacked_args, unpacked_arg_types

        nargs = target.nargs
        keys = list(target.args.keys())
        if implicit_receiver:
            nargs -= 1
            keys = keys[1:]
        args = unpacked_args[:nargs]
        arg_types = unpacked_arg_types[:nargs]

        var_args = unpacked_args[nargs:]
        var_arg_types = unpacked_arg_types[nargs:]

        no_missing = len(keys) - len(args)
        args += [False] * no_missing
        arg_types += [False] * no_missing

        kw_args = OrderedDict()

        # Named args
        for kw in keywords:
            if kw.arg in keys:
                index = keys.index(kw.arg)
                arg_stmt, arg_expr = self.translate_expr(kw.value, ctx)
                arg_type = self.get_type(kw.value, ctx)
                arg_stmts += arg_stmt
                args[index] = arg_expr
                arg_types[index] = arg_type
            else:
                if target.kw_arg:
                    kw_args[kw.arg] = kw.value

        # Default args
        for index, (arg, key) in enumerate(zip(args, keys)):
            if arg is False:
                # Not set yet, need default
                args[index] = target.args[key].default_expr
                assert args[index], '{} arg={}'.format(target.name, key)
                arg_types[index] = self.get_type(target.args[key].default, ctx)

        if target.var_arg:
            var_arg_list = self.create_tuple(var_args, var_arg_types, node, ctx)
            args.append(var_arg_list)
            arg_types.append(target.var_arg.type)

        if target.kw_arg:
            kw_stmt, kw_arg_dict = self._wrap_kw_args(kw_args, node,
                                                      target.kw_arg.type, ctx)
            args.append(kw_arg_dict)
            arg_types.append(target.kw_arg.type)
            arg_stmts += kw_stmt
        assert all(args), "Args translated into None: {}.".format(args)
        return arg_stmts, args, arg_types

    def _translate_receiver(self, node: ast.Call, target: PythonMethod,
            ctx: Context) -> Tuple[List[Stmt], List[Expr], List[PythonType]]:
        rec_stmts, receiver = self.translate_expr(node.func.value, ctx)
        receiver_type = self.get_type(node.func.value, ctx)
        if (target.method_type == MethodType.class_method and
                receiver_type.name != 'type'):
            receiver = self.type_factory.typeof(receiver, ctx)

        return rec_stmts, [receiver], [receiver_type]

    def _wrap_kw_args(self, args: Dict[str, ast.AST], node: ast.Call,
                      kw_type: PythonType, ctx: Context) -> StmtsAndExpr:
        """
        Wraps the given arguments into a dict to be passed to an **kwargs param.
        """
        res_var = ctx.current_function.create_variable('kw_args',
            ctx.module.global_module.classes[DICT_TYPE], self.translator)
        dict_class = ctx.module.global_module.classes[DICT_TYPE]
        constr_call = self.get_method_call(dict_class, '__init__', [],
                                           [], [res_var.ref()], node, ctx)
        position = self.to_position(node, ctx)
        type_inhale = self.viper.Inhale(self.type_check(res_var.ref(), kw_type,
                                                        position, ctx),
                                        position, self.no_info(ctx))
        stmt = constr_call + [type_inhale]
        str_type = ctx.module.global_module.classes[STRING_TYPE]
        for key, val in args.items():
            # Key string literal
            length = len(key)
            length_arg = self.viper.IntLit(length, self.no_position(ctx),
                                           self.no_info(ctx))
            val_arg = self.viper.IntLit(self._get_string_value(key),
                                        self.no_position(ctx),
                                        self.no_info(ctx))
            str_create_args = [length_arg, val_arg]
            str_create_arg_types = [None, None]
            func_name = '__create__'
            key_val = self.get_function_call(str_type, func_name,
                                             str_create_args,
                                             str_create_arg_types, node, ctx)
            val_stmt, val_val = self.translate_expr(val, ctx)
            val_type = self.get_type(val, ctx)
            args = [res_var.ref(), key_val, val_val]
            arg_types = [None, str_type, val_type]
            append_call = self.get_method_call(dict_class, '__setitem__', args,
                                               arg_types, [], node, ctx)
            stmt += val_stmt + append_call
        return stmt, res_var.ref()

    def inline_method(self, method: PythonMethod, args: List[PythonVar],
                      result_var: PythonVar, error_var: PythonVar,
                      ctx: Context) -> Tuple[List[Stmt], 'silver.ast.Label']:
        """
        Inlines a call to the given method, if the given argument vars contain
        the values of the arguments. Saves the result in result_var and any
        uncaught exceptions in error_var.
        """
        old_label_aliases = ctx.label_aliases.copy()
        old_var_aliases = ctx.var_aliases
        var_aliases = {}

        for name, arg in zip(method.args.keys(), args):
            var_aliases[name] = arg

        var_aliases[RESULT_NAME] = result_var
        if error_var:
            var_aliases[ERROR_NAME] = error_var
        ctx.inlined_calls.append(method)
        ctx.var_aliases = var_aliases
        ctx.label_aliases = {}
        ctx.ignore_waitlevel_constraints = True

        # Create local var aliases
        locals_to_copy = method.locals.copy()
        for local_name, local in locals_to_copy.items():
            local_var = ctx.current_function.create_variable(local_name,
                                                             local.type,
                                                             self.translator)
            ctx.set_alias(local_name, local_var, local)

        # Create label aliases
        for label in method.labels:
            new_label = ctx.current_function.get_fresh_name(label)
            ctx.label_aliases[label] = new_label
        end_label_name = ctx.label_aliases[END_LABEL]
        end_label = self.viper.Label(end_label_name, self.no_position(ctx),
                                     self.no_info(ctx))
        ctx.added_handlers.append((method, ctx.var_aliases, ctx.label_aliases))

        # Translate body
        start, end = get_body_indices(method.node.body)
        stmts = []

        for stmt in method.node.body[start:end]:
            stmts += self.translate_stmt(stmt, ctx)

        ctx.inlined_calls.remove(method)
        ctx.var_aliases = old_var_aliases
        ctx.label_aliases = old_label_aliases
        return stmts, end_label

    def _inline_call(self, method: PythonMethod, node: ast.Call, is_super: bool,
                     inline_reason: str, ctx: Context) -> StmtsAndExpr:
        """
        Inlines a statically bound call to the given method. If is_super is set,
        adds self to the arguments, since this will not be part of the args
        of the call node.
        """
        assert ctx.current_function
        if method in ctx.inlined_calls:
            raise InvalidProgramException(node, 'recursive.static.call')
        position = self.to_position(node, ctx)
        old_position = ctx.position
        ctx.position.append((inline_reason, position))
        arg_stmts, arg_vals, arg_types = self._translate_call_args(node, ctx)
        args = []
        stmts = arg_stmts

        # Create local vars for parameters and assign args to them
        if is_super:
            arg_vals = ([next(iter(ctx.actual_function.args.values())).ref()] +
                        arg_vals)
        for arg_val, (_, arg) in zip(arg_vals, method.args.items()):
            arg_var = ctx.current_function.create_variable('arg', arg.type,
                                                           self.translator)
            assign = self.viper.LocalVarAssign(arg_var.ref(), arg_val,
                                               self.to_position(node, ctx),
                                               self.no_info(ctx))
            stmts.append(assign)
            args.append(arg_var)

        # Create target vars
        res_var = None
        if method.type:
            res_var = ctx.current_function.create_variable(RESULT_NAME,
                                                           method.type,
                                                           self.translator)
        optional_error_var = None
        error_var = self.get_error_var(node, ctx)
        if method.declared_exceptions:
            var = PythonVar(ERROR_NAME, None,
                            ctx.module.global_module.classes['Exception'])
            var._ref = error_var
            optional_error_var = var
        old_fold = ctx.ignore_family_folds
        ctx.ignore_family_folds = True
        inline_stmts, end_lbl = self.inline_method(method, args, res_var,
                                                   optional_error_var, ctx)
        ctx.ignore_family_folds = old_fold
        stmts += inline_stmts
        if end_lbl:
            stmts.append(end_lbl)
        if method.declared_exceptions:
            stmts += self.create_exception_catchers(error_var,
                ctx.actual_function.try_blocks, node, ctx)
        # Return result
        result = res_var.ref() if method.type else None
        ctx.position.pop()
        return stmts, result

    def translate_call_in_union(self, arg_stmts: List[Stmt], args: List[Expr],
                                arg_types: List[PythonType], rectype: UnionType,
                                node: ast.Call, ctx: Context,
                                impure: bool) -> StmtsAndExpr:
        """
        Translate a method call or function call when the receiver is of type
        union. A chain of if-then-else statements or expressions will be used
        to call the method or function of each class in the union.
        """
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        guard_stmts_expr = []
        value_returned = False
        pure = True

        # For each type in union (subclasses first)
        for type in toposort_classes(rectype.get_types() - {None}):

            # Create guard checking if receiver is an instance of this type
            guard = self.var_type_check(node.func.value.id, type, pos, ctx)

            # Translate the method call (method of this type)
            stmts, expr = self.translate_normal_call(
                type.get_func_or_method(node.func.attr), arg_stmts, args,
                arg_types, node, ctx, impure)

            # Stores guard and translated method call as tuple
            guard_stmts_expr.append((guard, stmts, expr))

            # Calculate properties about the return value
            value_returned = value_returned or expr is not None
            pure = pure and len(stmts) == 0

        # Sanitize return value properties
        assert not pure or value_returned

        if not pure and value_returned:

            # Create a variable to assign the returned value
            return_var = ctx.current_function.create_variable('return_var',
                         ctx.module.global_module.classes[OBJECT_TYPE],
                         self.translator)

            for _, stmts, expr in guard_stmts_expr:
                if expr is not None:

                    # Assign the returned value to fresh variable
                    stmts.append(self.viper.LocalVarAssign(return_var.ref(
                                 node, ctx), expr, pos, info))

        if pure:

            # Chain list of guard and pure expression tuples in an if-then-else
            # expression
            guarded_expr = [(guard, expr) for guard, _, expr in guard_stmts_expr]
            return [], chain_cond_exp(guarded_expr, self.viper, pos, info, ctx)
        else:

            # Chain guards and blocks in an if-then-else statement
            guarded_blocks = [(guard, self.translate_block(stmts, pos, info))
                              for guard, stmts, _ in guard_stmts_expr]
            return ([chain_if_stmts(guarded_blocks, self.viper, pos, info, ctx)],
                    return_var._ref if value_returned else None)

    def translate_normal_call_node(self, node: ast.Call, ctx: Context,
                                   impure=False) -> StmtsAndExpr:
        """
        Translates a call node which refers to a 'normal' function, method or predicate.
        """
        arg_stmts, args, arg_types = self._translate_call_args(node, ctx)
        target = self._get_call_target(node, ctx)
        if not target:
            # Handle method calls when receiver's type is union
            if isinstance(node.func, ast.Attribute):
                rectype = self.get_type(node.func.value, ctx)
                if (isinstance(rectype, UnionType) and
                        not isinstance(rectype, OptionalType)):
                    # Check if the call refers to a predicate (family),
                    # in which case we don't need special treatment.
                    target_pred = rectype.python_class.get_predicate(node.func.attr)
                    if target_pred:
                        return self.translate_normal_call(target_pred, arg_stmts, args,
                                                          arg_types, node, ctx, impure)
                    # Otherwise apply special union treatment.
                    return self.translate_call_in_union(arg_stmts, args, arg_types,
                                                        rectype, node, ctx, impure)

            # Must be a function that exists (otherwise mypy would complain)
            # we don't know, so probably some builtin we don't support yet.
            msg = 'Unsupported builtin function'
            if ctx.actual_function.method_type == MethodType.class_method:
                msg += ' or indirect call of classmethod argument'
            raise UnsupportedException(node, msg + '.')
        if isinstance(target, PythonClass):
            return self.translate_constructor_call(target, node, args,
                                                   arg_stmts, ctx)
        return self.translate_normal_call(target, arg_stmts, args, arg_types, node, ctx,
                                          impure)

    def translate_normal_call(self, target: PythonMethod, arg_stmts: List[Stmt],
                              args: List[Expr], arg_types: List[PythonType],
                              node: ast.AST, ctx: Context, impure=False) -> StmtsAndExpr:
        """
        Translates a 'normal' function call, i.e., target must refer to a normal function,
        method or predicate.
        """
        formal_args = []
        name = get_func_name(node)
        position = self.to_position(node, ctx)
        is_predicate = True
        if isinstance(node.func, ast.Attribute):
            receiver_target = self.get_target(node.func.value, ctx)
            if (isinstance(receiver_target, PythonClass) and
                    (not isinstance(node.func.value, (ast.Call, ast.Str)) or
                             get_func_name(node.func.value) == 'super')):
                if target.method_type == MethodType.static_method:
                    # Static method
                    receiver_class = None
                    is_predicate = target.predicate
                elif target.method_type == MethodType.class_method:
                    rec_stmt, receiver = self.translate_expr(node.func.value,
                                                             ctx)
                    arg_stmts = rec_stmt + arg_stmts
                    args = [receiver] + args
                    arg_types = ([ctx.module.global_module.classes['type']] +
                                 arg_types)
                    receiver_class = receiver_target
                    is_predicate = False
                else:
                    # Statically bound call
                    is_super = get_func_name(node.func.value) == 'super'
                    if is_super:
                        if not self.is_valid_super_call(node.func.value,
                                                        ctx.actual_function):
                            raise InvalidProgramException(node.func.value,
                                                          'invalid.super.call')
                    return self._inline_call(target, node, is_super,
                                            'static call', ctx)
            elif isinstance(receiver_target, PythonModule):
                # Normal, receiverless call to imported function
                receiver_class = None
                is_predicate = target.predicate
            elif (isinstance(node.func.value, ast.Call) and
                        get_func_name(node.func.value) == 'super'):
                    # Super call
                    return self._inline_call(target, node, True, 'static call',
                                             ctx)
            else:
                # Method called on an object
                recv_stmts, recv_exprs, recv_types = self._translate_receiver(
                    node, target, ctx)
                is_predicate = target.predicate
                receiver_class = target.cls
                if target.method_type != MethodType.static_method:
                    arg_stmts = recv_stmts + arg_stmts
                    args = recv_exprs + args
                    arg_types = recv_types + arg_types
        else:
            # Global function/method called
            receiver_class = None
            is_predicate = target.predicate
        actual_args = []
        target_params = target.get_args()
        if target.var_arg:
            target_params.append(target.var_arg)
        if target.kw_arg:
            target_params.append(target.kw_arg)
        for arg, param, type in zip(args, target_params, arg_types):
            target_type = self.translate_type(param.type, ctx)
            actual_args.append(self.convert_to_type(arg, target_type, ctx))
        args = actual_args
        for arg in target.get_args():
            formal_args.append(arg.decl)
        target_name = target.sil_name
        if is_predicate:
            if receiver_class:
                family_root = receiver_class
                while (family_root.superclass and
                       family_root.superclass.get_predicate(name)):
                    family_root = family_root.superclass
                target_name = family_root.get_predicate(name).sil_name
            perm = self.viper.FullPerm(position, self.no_info(ctx))
            if not impure:
                raise InvalidProgramException(node, 'invalid.contract.position')
            return arg_stmts, self.create_predicate_access(target_name, args,
                                                           perm, node, ctx)
        elif target.pure:
            return self._translate_function_call(target, args, formal_args,
                                                 arg_stmts, position, node, ctx)
        else:
            return self._translate_method_call(target, args, arg_stmts,
                                               position, node, ctx)

    def _is_thread_method_call(self, node: ast.Call, name: str, ctx: Context) -> bool:
        """
        Checks if a given call calls a method of the given name defined by the Thread
        class.
        """
        if isinstance(node.func, ast.Attribute):
            val_type = self.get_type(node.func.value, ctx)
            if (isinstance(val_type, PythonClass) and
                    val_type.name == "Thread" and
                    node.func.attr == name):
                return True
        return False

    def translate_Call(self, node: ast.Call, ctx: Context, impure=False,
                       statement=False) -> StmtsAndExpr:
        """
        Translates any kind of call. This can be a call to a contract function
        like Assert, a builtin Python function like isinstance, a
        constructor call, a 'call' to a predicate, a pure function or impure
        method call, on a receiver object or not.
        Top-level contract statements like Assert are only allowed if the
        'statement' flag is set.
        """

        is_name = isinstance(node.func, ast.Name)
        func_name = get_func_name(node)
        if is_name:
            if func_name in CONTRACT_WRAPPER_FUNCS:
                raise InvalidProgramException(node, 'invalid.contract.position')
            elif func_name in CONTRACT_FUNCS:
                return self.translate_contractfunc_call(node, ctx, impure, statement)
            elif func_name in IO_CONTRACT_FUNCS:
                return self.translate_io_contractfunc_call(node, ctx)
            elif func_name in OBLIGATION_CONTRACT_FUNCS:
                return self.translate_obligation_contractfunc_call(node, ctx)
            elif func_name in BUILTINS:
                return self._translate_builtin_func(node, ctx)
            elif func_name == "Thread":
                return self._translate_thread_creation(node, ctx)
            elif func_name in BUILTIN_PREDICATES:
                return self.translate_contractfunc_call(node, ctx, impure)
        if self._is_cls_call(node, ctx):
            return self._translate_cls_call(node, ctx)
        elif isinstance(self.get_target(node, ctx), PythonIOOperation):
            return self.translate_io_operation_call(node, ctx)
        elif self._is_thread_method_call(node, 'start', ctx):
            return self._translate_thread_start(node, ctx)
        elif self._is_thread_method_call(node, 'join', ctx):
            return self._translate_thread_join(node, ctx)
        else:
            return self.translate_normal_call_node(node, ctx, impure)

    def _get_arg(self, nargs: int, keywords: List[str], index: int, kw: str) -> int:
        if nargs > index:
            return index
        if kw in keywords:
            return nargs + keywords.index(kw)
        return None

    def _handle_thread_constructor_args(self, node: ast.Call, ctx: Context):
        pos, info = self.to_position(node, ctx), self.no_info(ctx)
        # Map arguments to parameters
        arg_exprs = [a for a in node.args] + [kw.value for kw in node.keywords]
        keywords = [kw.arg for kw in node.keywords]

        group_arg = self._get_arg(len(node.args), keywords, 0, 'group')
        target_arg = self._get_arg(len(node.args), keywords, 1, 'target')
        name_arg = self._get_arg(len(node.args), keywords, 2, 'name')
        args_arg = self._get_arg(len(node.args), keywords, 3, 'args')
        kwargs_arg = self._get_arg(len(node.args), keywords, 4, 'kwargs')
        daemon_arg = self._get_arg(len(node.args), keywords, 5, 'daemon')

        if target_arg is None:
            raise InvalidProgramException(node, 'invalid.thread.creation')
        if kwargs_arg or daemon_arg:
            raise UnsupportedException(node, 'Unsupported thread parameter.')

        # Translate argument expressions, except target and args.
        thread_arg_stmts = []
        thread_arg_vals = []
        for index, expr in enumerate(arg_exprs):
            if index is target_arg or index is args_arg:
                thread_arg_vals.append(None)
                continue
            arg_stmt, arg_val = self.translate_expr(expr, ctx)
            thread_arg_stmts.extend(arg_stmt)
            thread_arg_vals.append(arg_val)

        target_arg = arg_exprs[target_arg]
        args_arg = arg_exprs[args_arg]

        if group_arg is not None:
            # The group argument must be None, everything else leads to a runtime error.
            group_none_pos = self.to_position(node, ctx, error_string='group is None',
                                              rules=rules.THREAD_CREATION_GROUP_NONE)
            null = self.viper.NullLit(group_none_pos, info)
            is_none = self.viper.EqCmp(thread_arg_vals[group_arg], null, group_none_pos,
                                       info)
            assert_none = self.viper.Assert(is_none, group_none_pos, info)
            thread_arg_stmts.append(assert_none)

        target = self.get_target(target_arg, ctx)
        if not isinstance(target, PythonMethod):
            raise InvalidProgramException(node, 'invalid.thread.creation')
        if target.pure or target.predicate:
            raise InvalidProgramException(node, 'invalid.thread.creation')
        if not isinstance(args_arg, ast.Tuple):
            raise InvalidProgramException(node, 'invalid.thread.creation')
        meth_args = args_arg.elts if args_arg is not None else []

        if isinstance(target_arg, ast.Attribute):
            receiver = self.get_target(target_arg.value, ctx)
            if not isinstance(receiver, PythonType):
                meth_args = [target_arg.value] + meth_args

        if len(meth_args) != len(target.args):
            no_default_args = [arg for arg in target.args.values()
                               if arg.default_expr is None]
            if len(no_default_args) != len(target.args):
                raise UnsupportedException(node, 'Thread target with default arguments.')
            else:
                raise InvalidProgramException(node, 'invalid.thread.creation')
        return thread_arg_stmts, meth_args, target

    def _translate_thread_creation(self, node: ast.Call,
                                   ctx: Context) -> StmtsAndExpr:
        """Translates the instantiation of a Thread object."""
        pos, info = self.to_position(node, ctx), self.no_info(ctx)

        thread_arg_stmts, meth_args, target = self._handle_thread_constructor_args(node,
                                                                                   ctx)

        # Create thread object
        thread_class = ctx.module.global_module.classes['Thread']
        thread_var = ctx.actual_function.create_variable('threadingVar', thread_class,
                                                         self.translator)
        thread = thread_var.ref(node, ctx)
        newstmt = self.viper.NewStmt(thread, [], pos, info)

        thread_object_type = self.type_check(thread, thread_class, pos, ctx)
        inhale_thread_type = self.viper.Inhale(thread_object_type, pos, info)

        # Inhale MayStart(t)
        start_pred_acc = self.viper.PredicateAccess([thread], THREAD_START_PRED,
                                                    pos, info)
        full_perm = self.viper.FullPerm(pos, info)
        start_pred = self.viper.PredicateAccessPredicate(start_pred_acc, full_perm, pos,
                                                         info)
        inhale_start_perm = self.viper.Inhale(start_pred, pos, info)

        # Check that given arguments match target method's parameter types,
        # and associate arguments and target method with thread object (by inhaling
        # getArg and getMethod information).
        arg_stmts = []
        arg_assumptions = self.viper.TrueLit(pos, info)
        arg_type_checks = self.viper.TrueLit(pos, info)
        method_args = list(target._args.values())
        for i, arg in enumerate(meth_args):
            arg_stmt, arg_val = self.translate_expr(arg, ctx, self.viper.Ref)
            arg_stmts.extend(arg_stmt)
            index = self.viper.IntLit(i, pos, info)
            arg_func = self.viper.DomainFuncApp(GET_ARG_FUNC, [thread, index],
                                                self.viper.Ref, pos, info, THREAD_DOMAIN)
            func_equal = self.viper.EqCmp(arg_func, arg_val, pos, info)
            arg_assumptions = self.viper.And(arg_assumptions, func_equal, pos, info)
            arg_type_check = self.type_check(arg_val, method_args[i].type, pos, ctx)
            arg_type_checks = self.viper.EqCmp(arg_type_checks, arg_type_check, pos, info)

        method_id_type = self.viper.DomainType(METHOD_ID_DOMAIN, {}, [])
        thread_method = self.viper.DomainFuncApp(GET_METHOD_FUNC, [thread],
                                                 method_id_type, pos, info, THREAD_DOMAIN)
        actual_method = self.viper.DomainFuncApp(target.threading_id, [], method_id_type,
                                                 pos, info, METHOD_ID_DOMAIN)
        inhale_method = self.viper.Inhale(self.viper.EqCmp(thread_method, actual_method,
                                                           pos, info),
                                          pos, info)
        arg_check_pos = self.to_position(node, ctx, rules=rules.THREAD_CREATION_ARG_TYPE)
        check_arg_types = self.viper.Assert(arg_type_checks, arg_check_pos, info)
        inhale_args = self.viper.Inhale(arg_assumptions, pos, info)
        stmts = thread_arg_stmts + arg_stmts + [newstmt, inhale_thread_type,
                                                inhale_method, check_arg_types,
                                                inhale_args, inhale_start_perm]
        return stmts, thread

    def _translate_thread_start(self, node: ast.Call,
                                ctx: Context) -> StmtsAndExpr:
        """Translates a thread start call."""
        pos, info = self.to_position(node, ctx), self.no_info(ctx)
        assert isinstance(node.func, ast.Attribute)
        thread_stmt, thread = self.translate_expr(node.func.value, ctx)
        ctx.current_thread_object, ctx.is_thread_start = thread, True

        # Resolve list of possible target methods.
        method_options = []
        for arg in node.args:
            target = self.get_target(arg, ctx)
            if not (isinstance(target, PythonMethod) and not target.pure and
                        not target.predicate):
                raise InvalidProgramException(node, 'invalid.thread.start')
            method_options.append(target)

        stmts = []

        # Exhale MayStart predicate
        start_pred_pos = self.to_position(node, ctx, rules=rules.THREAD_START_PERMISSION)
        full_perm = self.viper.FullPerm(start_pred_pos, info)
        start_pred_acc = self.viper.PredicateAccess([thread], THREAD_START_PRED,
                                                    start_pred_pos,
                                                    info)
        start_pred = self.viper.PredicateAccessPredicate(start_pred_acc, full_perm,
                                                         start_pred_pos, info)
        stmts.append(self.viper.Exhale(start_pred, start_pred_pos, info))

        # Assert that actual target method is in list of options.
        options_pos = self.to_position(node, ctx,
                                       rules=rules.THREAD_START_METHOD_UNLISTED)
        correct_method = self.viper.FalseLit(options_pos, info)
        method_id_type = self.viper.DomainType(METHOD_ID_DOMAIN, {}, [])
        actual_method = self.viper.DomainFuncApp(GET_METHOD_FUNC, [thread],
                                                 method_id_type, options_pos, info,
                                                 THREAD_DOMAIN)
        for method in method_options:
            this_method = self.viper.DomainFuncApp(method.threading_id, [],
                                                   method_id_type, options_pos, info,
                                                   METHOD_ID_DOMAIN)
            this_option = self.viper.EqCmp(actual_method, this_method, options_pos, info)
            correct_method = self.viper.Or(correct_method, this_option, options_pos, info)
        stmts.append(self.viper.Assert(correct_method, options_pos, info))
        precond_pos = self.to_position(node, ctx, rules=rules.THREAD_START_PRECONDITION)

        # Actual fork operation is carried out elsewhere.
        stmts.extend(self.create_method_fork(ctx, method_options, thread, precond_pos,
                                             info, node))
        ctx.current_thread_object, ctx.is_thread_start = None, False
        return thread_stmt + stmts, None

    def _translate_thread_join(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translates a thread join call."""

        pos, info = self.to_position(node, ctx), self.no_info(ctx)
        assert isinstance(node.func, ast.Attribute)
        thread_stmt, thread = self.translate_expr(node.func.value, ctx)
        ctx.current_thread_object, ctx.is_thread_start = thread, False
        stmts = thread_stmt

        # Assert that thread may be joined.
        joinable_pos = self.to_position(node, ctx, rules=rules.THREAD_JOIN_JOINABLE)
        joinable_func = self.viper.FuncApp(JOINABLE_FUNC, [thread],
                                           joinable_pos, info,
                                           self.viper.Bool)

        stmts.append(self.viper.Assert(joinable_func, joinable_pos, info))

        # Check that thread wait level is above current wait level (prevents deadlocks).
        wait_level_pos = self.to_position(node, ctx, rules=rules.THREAD_JOIN_WAITLEVEL)
        thread_level = self.create_level_call(sil.RefExpr(thread))
        residue_var = sil.PermVar(ctx.actual_function.obligation_info.residue_level)
        obligation_assertion = self.create_level_below(thread_level, residue_var, ctx)
        obligation_assertion = obligation_assertion.translate(self, ctx, wait_level_pos,
                                                              info)
        stmts.append(self.viper.Assert(obligation_assertion, wait_level_pos, info))

        # Check how much permission is held to ThreadPost. This amount of the thread
        # postcondition will be inhaled later.
        post_pred_acc = self.viper.PredicateAccess([thread], THREAD_POST_PRED, pos, info)
        post_perm = self.viper.CurrentPerm(post_pred_acc, pos, info)
        any_perm = self.viper.PermGtCmp(post_perm, self.viper.NoPerm(pos, info), pos,
                                        info)

        # Any subsequently translated assertions (postconditions) are multiplied by this
        # amount of permission.
        ctx.perm_factor = post_perm

        object_class = ctx.module.global_module.classes[OBJECT_TYPE]
        res_var = ctx.actual_function.create_variable('join_result', object_class,
                                                      self.translator)

        # Resolve list of possible thread target methods.
        method_options = []
        for arg in node.args:
            target = self.get_target(arg, ctx)
            if not (isinstance(target, PythonMethod) and not target.pure and
                        not target.predicate):
                raise InvalidProgramException(node, 'invalid.thread.join')
            method_options.append(target)

        # Conditionally inhale postconditions of target methods.
        for method in method_options:
            stmts.append(self._inhale_possible_thread_post(method, thread, res_var,
                                                           any_perm, pos, ctx))

        ctx.perm_factor = None

        post_pred = self.viper.PredicateAccessPredicate(post_pred_acc, post_perm, pos,
                                                        info)
        exhale_pred = self.viper.Exhale(post_pred, pos, info)
        stmts.append(exhale_pred)
        ctx.current_thread_object, ctx.is_thread_start = None, False

        return stmts, None

    def _inhale_possible_thread_post(self, method: PythonMethod, thread: Expr,
                                     res_var: PythonVar, any_perm: Expr, pos: Position,
                                     ctx: Context) -> Stmt:
        """
        Creates a statement that inhales this method's postcondition if this is
        the target method of the given thread.
        """
        method_stmts = []
        info = self.no_info(ctx)
        else_block = self.translate_block([], pos, info)
        method_id_type = self.viper.DomainType(METHOD_ID_DOMAIN, {}, [])
        actual_method = self.viper.DomainFuncApp(GET_METHOD_FUNC, [thread],
                                                 method_id_type, pos, info, THREAD_DOMAIN)

        # Set arg aliases with types
        for index, arg in enumerate(method._args.values()):
            arg_var = ctx.actual_function.create_variable('thread_arg', arg.type,
                                                          self.translator)
            ctx.set_alias(arg.name, arg_var)
            id = self.viper.IntLit(index, pos, info)
            arg_func = self.viper.DomainFuncApp(GET_ARG_FUNC, [thread, id],
                                                self.viper.Ref, pos, info,
                                                THREAD_DOMAIN)
            method_stmts.append(self.viper.LocalVarAssign(arg_var.ref(), arg_func,
                                                          pos, info))
            method_stmts.append(self.viper.Inhale(self.type_check(arg_var.ref(),
                                                                  arg.type, pos, ctx,
                                                                  inhale_exhale=False),
                                                  pos, info))
        if method.type:
            ctx.set_alias(RESULT_NAME, res_var)
            res_var.type = method.type

        # Set old values
        collector = OldExpressionCollector()
        normalizer = OldExpressionTransformer()
        normalizer.arg_names = [arg for arg in method._args]
        for post, _ in method.postcondition:
            collector.visit(post)
        for old in collector.expressions:
            print_old = normalizer.visit(copy.deepcopy(old))
            key = pprint(print_old)
            id = self.viper.IntLit(self._get_string_value(key), pos, info)
            old_func = self.viper.DomainFuncApp(GET_OLD_FUNC, [thread, id],
                                                self.viper.Ref, pos, info,
                                                THREAD_DOMAIN)
            ctx.set_old_expr_alias(key, old_func)
            old_type = self.get_type(old, ctx)
            method_stmts.append(self.viper.Inhale(self.type_check(old_func,
                                                                  old_type, pos, ctx,
                                                                  inhale_exhale=False),
                                                  pos, info))

        post_assertion = self.viper.TrueLit(pos, info)
        ctx.inlined_calls.append(method)

        # Translate postcondition
        for post, _ in method.postcondition:
            _, post_val = self.translate_expr(post, ctx, target_type=self.viper.Bool,
                                              impure=True)
            post_assertion = self.viper.And(post_assertion, post_val, pos, info)

        ctx.inlined_calls.pop()
        ctx.clear_old_expr_aliases()
        for name in method._args:
            ctx.remove_alias(name)
        ctx.remove_alias(RESULT_NAME)
        ctx.ignore_waitlevel_constraints = False

        # Inhale postcondition if there's a permission to ThreadPost and this method
        # is the actual method.
        method_stmts.append(self.viper.Inhale(post_assertion, pos, info))
        then_block = self.translate_block(method_stmts, pos, info)
        this_method = self.viper.DomainFuncApp(method.threading_id, [],
                                               method_id_type, pos, info,
                                               METHOD_ID_DOMAIN)
        correct_method = self.viper.EqCmp(actual_method, this_method, pos, info)
        cond = self.viper.And(any_perm, correct_method, pos, info)
        return self.viper.If(cond, then_block, else_block, pos, info)

    def _is_cls_call(self, node: ast.Call, ctx: Context) -> bool:
        """
        Checks if the given call is a call to the cls parameter in a class
        method.
        """
        if (ctx.actual_function and
            isinstance(ctx.actual_function, PythonMethod) and
            ctx.actual_function.method_type == MethodType.class_method):
            if isinstance(node.func, ast.Name):
                if node.func.id == next(iter(ctx.actual_function.args.keys())):
                    return True
        return False

    def _translate_cls_call(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the cls parameter in a class method.
        """
        target_class = ctx.actual_function.cls
        args = []
        arg_stmts = []
        for arg in node.args:
            arg_stmt, arg_val = self.translate_expr(arg, ctx)
            arg_stmts += arg_stmt
            args.append(arg_val)
        res_var = ctx.current_function.create_variable(target_class.name +
                                                       '_res',
                                                       target_class,
                                                       self.translator)

        pos = self.to_position(node, ctx)
        fields = list(target_class.all_fields)
        may_set_inhales = [self.viper.Inhale(self.get_may_set_predicate(res_var.ref(),
                                                                        f, ctx),
                                             pos, self.no_info(ctx))
                           for f in fields]
        new = self.viper.NewStmt(res_var.ref(), [], self.no_position(ctx),
                                 self.no_info(ctx))

        type_stmt, dynamic_type = self.translate_expr(node.func, ctx)
        assert not type_stmt
        result_has_type = self.type_factory.dynamic_type_check(res_var.ref(),
            dynamic_type, self.to_position(node, ctx), ctx)
        # Inhale the type information about the newly created object
        # so that it's already present when calling __init__.
        type_inhale = self.viper.Inhale(result_has_type, pos,
                                        self.no_info(ctx))
        args = [res_var.ref()] + args
        stmts = [new, type_inhale] + may_set_inhales
        target = target_class.get_method('__init__')
        if target:
            target_class = target.cls
            targets = []
            if target.declared_exceptions:
                error_var = self.get_error_var(node, ctx)
                targets.append(error_var)
            target_method = target_class.get_method('__init__')
            method_name = target_method.sil_name
            init = self.create_method_call_node(ctx, method_name, args, targets,
                                                self.to_position(node, ctx),
                                                self.no_info(ctx),
                                                target_method=target_method,
                                                target_node=node)
            stmts.extend(init)
            if target.declared_exceptions:
                catchers = self.create_exception_catchers(error_var,
                    ctx.actual_function.try_blocks, node, ctx)
                stmts = stmts + catchers
        return arg_stmts + stmts, res_var.ref()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Silver expressions."""


from nagini_translation.lib.program_nodes import (
    PythonVar,
)
from nagini_translation.lib.silver_nodes.base import Expression as IExpr
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)


class Expression(IExpr):   # pylint: disable=abstract-method
    """A base class for all expressions."""

    EqCmp = None    # To avoid cyclic imports.
    NeCmp = None

    def __eq__(self, other) -> 'Expression':
        return Expression.EqCmp(self, other)

    def __ne__(self, other) -> 'Expression':
        return Expression.NeCmp(self, other)


class VarDecl(Expression):
    """A variable declaration."""

    def __init__(self, var: PythonVar) -> None:
        self._var = var

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return self._var.decl


class AnyVar(Expression):
    """A variable of a unknown type."""

    def __init__(self, var: PythonVar) -> None:
        self._var = var

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return self._var.ref()


__all__ = (
    'AnyVar',
    'Expression',
    'VarDecl',
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
import copy

from nagini_translation.lib.constants import BOOL_TYPE
from nagini_translation.lib.util import InvalidProgramException
from nagini_translation.lib.program_nodes import PythonMethod
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator
from toposort import toposort_flatten
from typing import List


class PredicateTranslator(CommonTranslator):

    def translate_predicate(self, pred: PythonMethod,
                            ctx: Context) -> 'ast.silver.Predicate':
        """
        Translates pred to a Silver predicate.
        """
        if not pred.type or pred.type.name != BOOL_TYPE:
            raise InvalidProgramException(pred.node, 'invalid.predicate')
        assert ctx.current_function is None
        ctx.current_function = pred
        args = []
        arg_types = self.viper.TrueLit(self.no_position(ctx), self.no_info(ctx))
        for name, arg in pred.args.items():
            args.append(arg.decl)
            arg_type = self.type_check(arg.ref(), arg.type,
                                       self.no_position(ctx), ctx)
            arg_types = self.viper.And(arg_types, arg_type,
                                       self.no_position(ctx), self.no_info(ctx))
        if len(pred.node.body) != 1:
            raise InvalidProgramException(pred.node,
                                          'invalid.predicate')

        content = pred.node.body[0]
        if isinstance(content, ast.Return):
            content = content.value
        stmt, body = self.translate_expr(
            content,
            ctx, impure=True,
            target_type=self.viper.Bool)
        if stmt:
            raise InvalidProgramException(pred.node,
                                          'invalid.predicate')
        body = self.viper.And(arg_types, body, self.no_position(ctx),
                              self.no_info(ctx))
        ctx.current_function = None
        return self.viper.Predicate(pred.sil_name, args, body,
                                    self.to_position(pred.node, ctx),
                                    self.no_info(ctx))

    def translate_predicate_family(self, root: PythonMethod,
                                   preds: List[PythonMethod],
                                   ctx: Context) -> 'ast.silver.Predicate':
        """
        Translates the methods in preds, whose root (which they all override)
        is root, to a family-predicate in Silver.
        """
        dependencies = {}
        for pred in preds:
            value = {pred.overrides} if pred.overrides else set()
            dependencies[pred] = value
        sorted = toposort_flatten(dependencies, False)

        name = root.sil_name
        args = []
        self_var_ref = root.args[next(iter(root.args))].ref()
        arg_types = self.viper.TrueLit(self.no_position(ctx), self.no_info(ctx))
        self.bind_type_vars(root, ctx)
        for arg in root.args.values():
            args.append(arg.decl)
            arg_type = self.type_check(arg.ref(), arg.type,
                                       self.no_position(ctx), ctx)
            arg_types = self.viper.And(arg_types, arg_type,
                                       self.no_position(ctx), self.no_info(ctx))
        body = None
        assert not ctx.var_aliases
        for instance in sorted:
            ctx.var_aliases = {}
            assert not ctx.current_function
            if instance.type.name != BOOL_TYPE:
                raise InvalidProgramException(instance.node,
                                              'invalid.predicate')
            ctx.current_function = instance
            ctx.module = instance.module
            self.bind_type_vars(instance, ctx)
            # Replace variables in instance by variables in root, since we use the
            # parameter names from root.
            for root_name, current_name in zip(root.args.keys(),
                                               instance.args.keys()):
                root_var = root.args[root_name]
                # For the receiver parameter, we need it to have the same sil_name as
                # that of the root, but the type of the current instance when translating
                # it, otherwise some fields/functions/predicates may not be found.
                if root_name == next(iter(root.args.keys())):
                    root_var = copy.copy(root_var)
                    root_var.type = instance.cls
                ctx.set_alias(current_name, root_var)
            actual_body_start = 0
            while (actual_body_start < len(instance.node.body) and
                       isinstance(instance.node.body[actual_body_start], ast.Expr) and
                    isinstance(instance.node.body[actual_body_start].value, ast.Str)):
                actual_body_start += 1
            if len(instance.node.body[actual_body_start:]) != 1:
                raise InvalidProgramException(instance.node,
                                              'invalid.predicate')
            content = instance.node.body[actual_body_start]
            if isinstance(content, ast.Return):
                content = content.value

            stmt, current = self.translate_expr(
                    content,
                    ctx, impure=True,
                    target_type=self.viper.Bool)
            if stmt:
                raise InvalidProgramException(instance.node,
                                              'invalid.predicate')
            has_type = self.type_factory.type_check(self_var_ref, instance.cls,
                                                    self.to_position(
                                                        instance.node, ctx),
                                                    ctx)
            implication = self.viper.Implies(has_type, current,
                self.to_position(instance.node, ctx), self.no_info(ctx))
            ctx.current_function = None
            if body:
                body = self.viper.And(body, implication,
                    self.to_position(root.node, ctx), self.no_info(ctx))
            else:
                body = implication
        ctx.var_aliases = {}
        body = self.viper.And(arg_types, body, self.no_position(ctx),
                              self.no_info(ctx))
        return self.viper.Predicate(name, args, body,
                                    self.to_position(root.node, ctx),
                                    self.no_info(ctx))



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import types

from nagini_translation.lib.constants import FUNCTION_DOMAIN_NAME
from nagini_translation.lib.errors import error_manager, Rules


def getobject(package, name):
    return getattr(getattr(package, name + '$'), 'MODULE$')


class Function0:
    def apply(self):
        pass


LONG_SIZE = 2147483647


class ViperAST:
    """
    Provides convenient access to the classes which constitute the Viper AST.
    All constructors convert Python lists to Scala sequences, Python ints
    to Scala BigInts, and wrap Scala Option types where necessary.
    """

    def __init__(self, jvm, java, scala, viper, sourcefile):
        ast = viper.silver.ast
        self.ast = ast
        self.java = java
        self.scala = scala
        self.jvm = jvm
        self.nodes = {}
        self.used_names = set()
        self.used_names_sets = {}

        def getconst(name):
            return getobject(ast, name)
        self.QPs = getobject(ast.utility, 'QuantifiedPermissions')
        self.AddOp = getconst('AddOp')
        self.AndOp = getconst('AndOp')
        self.DivOp = getconst('DivOp')
        self.FracOp = getconst('FracOp')
        self.GeOp = getconst('GeOp')
        self.GtOp = getconst('GtOp')
        self.ImpliesOp = getconst('ImpliesOp')
        self.IntPermMulOp = getconst('IntPermMulOp')
        self.LeOp = getconst('LeOp')
        self.LtOp = getconst('LtOp')
        self.ModOp = getconst('ModOp')
        self.MulOp = getconst('MulOp')
        self.NegOp = getconst('NegOp')
        self.NotOp = getconst('NotOp')
        self.OrOp = getconst('OrOp')
        self.PermAddOp = getconst('PermAddOp')
        self.PermDivOp = getconst('PermDivOp')
        self.SubOp = getconst('SubOp')
        self.NoPosition = getconst('NoPosition')
        self.NoInfo = getconst('NoInfo')
        self.NoTrafos = getconst('NoTrafos')
        self.Int = getconst('Int')
        self.Bool = getconst('Bool')
        self.Ref = getconst('Ref')
        self.Perm = getconst('Perm')
        self.sourcefile = sourcefile
        self.none = getobject(scala, 'None')

    def is_available(self) -> bool:
        """
        Checks if the Viper AST is available, i.e., silver is on the Java classpath.
        """
        return self.jvm.is_known_class(self.ast.Program)

    def function_domain_type(self):
        return self.DomainType(FUNCTION_DOMAIN_NAME, {}, [])

    def empty_seq(self):
        return self.scala.collection.mutable.ListBuffer()

    def singleton_seq(self, element):
        result = self.scala.collection.mutable.ArraySeq(1)
        result.update(0, element)
        return result

    def append(self, list, to_append):
        if not to_append is None:
            lsttoappend = self.singleton_seq(to_append)
            list.append(lsttoappend)

    def to_seq(self, list):
        result = self.scala.collection.mutable.ArraySeq(len(list))
        for index in range(0, len(list)):
            result.update(index, list[index])
        return result.toList()

    def to_list(self, seq):
        result = []
        iterator = seq.toIterator()
        while iterator.hasNext():
            result.append(iterator.next())
        return result

    def to_map(self, dict):
        result = self.scala.collection.immutable.HashMap()
        for k, v in dict.items():
            result = result.updated(k, v)
        return result

    def to_big_int(self, num):
        # We cannot give integers directly to Scala if they don't
        # fit into a C long int, so we have to split things up.
        negative = num < 0
        if negative:
            num = -num
        cutoff = LONG_SIZE
        cutoff_int = self.java.math.BigInteger.valueOf(cutoff)
        rest = num
        result_int = self.java.math.BigInteger.valueOf(0)
        while rest > 0:
            current_part = rest % cutoff
            current_int = self.java.math.BigInteger.valueOf(current_part)
            result_int = result_int.multiply(cutoff_int)
            result_int = result_int.add(current_int)
            rest = rest // cutoff
        if negative:
            result_int = result_int.negate()
        return self.scala.math.BigInt(result_int)

    def Program(self, domains, fields, functions, predicates, methods, position, info):
        return self.ast.Program(self.to_seq(domains), self.to_seq(fields),
                                self.to_seq(functions), self.to_seq(predicates),
                                self.to_seq(methods), position, info, self.NoTrafos)

    def Function(self, name, args, type, pres, posts, body, position, info):
        body = self.scala.Some(body) if body is not None else self.none
        return self.ast.Function(name, self.to_seq(args), type,
                                 self.to_seq(pres),
                                 self.to_seq(posts), self.none,
                                 body, position, info, self.NoTrafos)

    def Method(self, name, args, returns, pres, posts, locals, body, position,
               info):
        if body is None:
            body_with_locals = self.none
        else:
            body_with_locals = self.scala.Some(self.Seqn([body], position, info, locals))
        method = getobject(self.ast, "MethodWithLabelsInScope")
        return method.apply(name, self.to_seq(args), self.to_seq(returns),
                            self.to_seq(pres), self.to_seq(posts),
                            body_with_locals, position, info,
                            self.NoTrafos)

    def Field(self, name, type, position, info):
        return self.ast.Field(name, type, position, info, self.NoTrafos)

    def Predicate(self, name, args, body, position, info):
        body = self.scala.Some(body) if body is not None else self.none
        return self.ast.Predicate(name, self.to_seq(args),
                                  body, position, info, self.NoTrafos)

    def PredicateAccess(self, args, pred_name, position, info):
        self.used_names.add(pred_name)
        return self.ast.PredicateAccess(self.to_seq(args), pred_name, position,
                                        info, self.NoTrafos)

    def PredicateAccessPredicate(self, loc, perm, position, info):
        return self.ast.PredicateAccessPredicate(loc, perm, position, info, self.NoTrafos)

    def Fold(self, predicate, position, info):
        return self.ast.Fold(predicate, position, info, self.NoTrafos)

    def Unfold(self, predicate, position, info):
        return self.ast.Unfold(predicate, position, info, self.NoTrafos)

    def Unfolding(self, predicate, expr, position, info):
        return self.ast.Unfolding(predicate, expr, position, info, self.NoTrafos)

    def SeqType(self, element_type):
        return self.ast.SeqType(element_type)

    def SetType(self, element_type):
        return self.ast.SetType(element_type)

    def MultisetType(self, element_type):
        return self.ast.MultisetType(element_type)

    def Domain(self, name, functions, axioms, typevars, position, info):
        return self.ast.Domain(name, self.to_seq(functions),
                               self.to_seq(axioms), self.to_seq(typevars),
                               position, info, self.NoTrafos)

    def DomainFunc(self, name, args, type, unique, position, info, domain_name):
        return self.ast.DomainFunc(name, self.to_seq(args), type, unique,
                                   position, info, domain_name, self.NoTrafos)

    def DomainAxiom(self, name, expr, position, info, domain_name):
        return self.ast.DomainAxiom(name, expr, position, info, domain_name,
                                    self.NoTrafos)

    def DomainType(self, name, type_vars_map, type_vars):
        map = self.to_map(type_vars_map)
        seq = self.to_seq(type_vars)
        return self.ast.DomainType(name, map,
                                   seq)

    def DomainFuncApp(self, func_name, args, type_passed,
                      position, info, domain_name, type_var_map={}):
        arg_decls = [self.LocalVarDecl('arg' + str(i), arg.typ(), arg.pos(),
                                       arg.info())
                     for i, arg in enumerate(args)]

        def type_passed_apply(slf):
            return type_passed

        def args_passed_apply(slf):
            return self.to_seq(arg_decls)

        type_passed_func = self.to_function0(type_passed_apply)
        args_passed_func = self.to_function0(args_passed_apply)
        result = self.ast.DomainFuncApp(func_name, self.to_seq(args),
                                        self.to_map(type_var_map), position,
                                        info, type_passed_func,
                                        args_passed_func, domain_name, self.NoTrafos)
        return result

    def TypeVar(self, name):
        return self.ast.TypeVar(name)

    def MethodCall(self, method_name, args, targets, position, info):
        self.used_names.add(method_name)
        return self.ast.MethodCall(method_name, self.to_seq(args),
                                   self.to_seq(targets), position, info, self.NoTrafos)

    def NewStmt(self, lhs, fields, position, info):
        return self.ast.NewStmt(lhs, self.to_seq(fields), position, info, self.NoTrafos)

    def Label(self, name, position, info):
        return self.ast.Label(name, self.to_seq([]), position, info, self.NoTrafos)

    def Goto(self, name, position, info):
        return self.ast.Goto(name, position, info, self.NoTrafos)

    def Seqn(self, body, position, info, locals=[]):
        return self.ast.Seqn(self.to_seq(body), self.to_seq(locals), position, info,
                             self.NoTrafos)

    def LocalVarAssign(self, lhs, rhs, position, info):
        return self.ast.LocalVarAssign(lhs, rhs, position, info, self.NoTrafos)

    def FieldAssign(self, lhs, rhs, position, info):
        return self.ast.FieldAssign(lhs, rhs, position, info, self.NoTrafos)

    def FieldAccess(self, receiver, field, position, info):
        return self.ast.FieldAccess(receiver, field, position, info, self.NoTrafos)

    def FieldAccessPredicate(self, fieldacc, perm, position, info):
        return self.ast.FieldAccessPredicate(fieldacc, perm, position, info, self.NoTrafos)

    def Old(self, expr, position, info):
        return self.ast.Old(expr, position, info, self.NoTrafos)

    def LabelledOld(self, expr, label, position, info):
        return self.ast.LabelledOld(expr, label, position, info, self.NoTrafos)

    def Inhale(self, expr, position, info):
        return self.ast.Inhale(expr, position, info, self.NoTrafos)

    def Exhale(self, expr, position, info):
        return self.ast.Exhale(expr, position, info, self.NoTrafos)

    def InhaleExhaleExp(self, inhale, exhale, position, info):
        return self.ast.InhaleExhaleExp(inhale, exhale, position, info, self.NoTrafos)

    def Assert(self, expr, position, info):
        return self.ast.Assert(expr, position, info, self.NoTrafos)

    def FullPerm(self, position, info):
        return self.ast.FullPerm(position, info, self.NoTrafos)

    def NoPerm(self, position, info):
        return self.ast.NoPerm(position, info, self.NoTrafos)

    def WildcardPerm(self, position, info):
        return self.ast.WildcardPerm(position, info, self.NoTrafos)

    def FractionalPerm(self, left, right, position, info):
        return self.ast.FractionalPerm(left, right, position, info, self.NoTrafos)

    def CurrentPerm(self, location, position, info):
        return self.ast.CurrentPerm(location, position, info, self.NoTrafos)

    def ForPerm(self, variable, access, body, position, info):
        if isinstance(access, self.ast.Predicate):
            self.used_names.add(access.name())
        variables = self.to_seq([variable])
        return self.ast.ForPerm(variables, access, body,
                                position, info, self.NoTrafos)

    def PermMinus(self, exp, position, info):
        return self.ast.PermMinus(exp, position, info, self.NoTrafos)

    def PermAdd(self, left, right, position, info):
        return self.ast.PermAdd(left, right, position, info, self.NoTrafos)

    def PermSub(self, left, right, position, info):
        return self.ast.PermSub(left, right, position, info, self.NoTrafos)

    def PermMul(self, left, right, position, info):
        return self.ast.PermMul(left, right, position, info, self.NoTrafos)

    def IntPermMul(self, left, right, position, info):
        return self.ast.IntPermMul(left, right, position, info, self.NoTrafos)

    def PermDiv(self, left, right, position, info):
        return self.ast.PermDiv(left, right, position, info, self.NoTrafos)

    def PermLtCmp(self, left, right, position, info):
        return self.ast.PermLtCmp(left, right, position, info, self.NoTrafos)

    def PermLeCmp(self, left, right, position, info):
        return self.ast.PermLeCmp(left, right, position, info, self.NoTrafos)

    def PermGtCmp(self, left, right, position, info):
        return self.ast.PermGtCmp(left, right, position, info, self.NoTrafos)

    def PermGeCmp(self, left, right, position, info):
        return self.ast.PermGeCmp(left, right, position, info, self.NoTrafos)

    def Not(self, expr, position, info):
        return self.ast.Not(expr, position, info, self.NoTrafos)

    def Minus(self, expr, position, info):
        return self.ast.Minus(expr, position, info, self.NoTrafos)

    def CondExp(self, cond, then, els, position, info):
        return self.ast.CondExp(cond, then, els, position, info, self.NoTrafos)

    def EqCmp(self, left, right, position, info):
        return self.ast.EqCmp(left, right, position, info, self.NoTrafos)

    def NeCmp(self, left, right, position, info):
        return self.ast.NeCmp(left, right, position, info, self.NoTrafos)

    def GtCmp(self, left, right, position, info):
        return self.ast.GtCmp(left, right, position, info, self.NoTrafos)

    def GeCmp(self, left, right, position, info):
        return self.ast.GeCmp(left, right, position, info, self.NoTrafos)

    def LtCmp(self, left, right, position, info):
        return self.ast.LtCmp(left, right, position, info, self.NoTrafos)

    def LeCmp(self, left, right, position, info):
        return self.ast.LeCmp(left, right, position, info, self.NoTrafos)

    def IntLit(self, num, position, info):
        return self.ast.IntLit(self.to_big_int(num), position, info, self.NoTrafos)

    def Implies(self, left, right, position, info):
        return self.ast.Implies(left, right, position, info, self.NoTrafos)

    def FuncApp(self, name, args, position, info, type, formalargs=None):
        self.used_names.add(name)
        if formalargs is None:
            formalargs = []
            for i, a in enumerate(args):
                formalargs.append(self.LocalVarDecl('p' + str(i), a.typ(), a.pos(),
                                                    a.info()))
        return self.ast.FuncApp(name, self.to_seq(args), position, info, type,
                                self.to_seq(formalargs), self.NoTrafos)

    def ExplicitSeq(self, elems, position, info):
        return self.ast.ExplicitSeq(self.to_seq(elems), position, info, self.NoTrafos)

    def ExplicitSet(self, elems, position, info):
        return self.ast.ExplicitSet(self.to_seq(elems), position, info, self.NoTrafos)

    def ExplicitMultiset(self, elems, position, info):
        return self.ast.ExplicitMultiset(self.to_seq(elems), position, info, self.NoTrafos)

    def EmptySeq(self, type, position, info):
        return self.ast.EmptySeq(type, position, info, self.NoTrafos)

    def EmptySet(self, type, position, info):
        return self.ast.EmptySet(type, position, info, self.NoTrafos)

    def EmptyMultiset(self, type, position, info):
        return self.ast.EmptyMultiset(type, position, info, self.NoTrafos)

    def LocalVarDecl(self, name, type, position, info):
        return self.ast.LocalVarDecl(name, type, position, info, self.NoTrafos)

    def LocalVar(self, name, type, position, info):
        return self.ast.LocalVar(name, type, position, info, self.NoTrafos)

    def Result(self, type, position, info):
        return self.ast.Result(type, position, info, self.NoTrafos)

    def AnySetContains(self, elem, s, position, info):
        return self.ast.AnySetContains(elem, s, position, info, self.NoTrafos)

    def AnySetUnion(self, left, right, position, info):
        return self.ast.AnySetUnion(left, right, position, info, self.NoTrafos)

    def AnySetSubset(self, left, right, position, info):
        return self.ast.AnySetSubset(left, right, position, info, self.NoTrafos)

    def SeqAppend(self, left, right, position, info):
        return self.ast.SeqAppend(left, right, position, info, self.NoTrafos)

    def SeqContains(self, elem, s, position, info):
        return self.ast.SeqContains(elem, s, position, info, self.NoTrafos)

    def SeqLength(self, s, position, info):
        return self.ast.SeqLength(s, position, info, self.NoTrafos)

    def SeqIndex(self, s, ind, position, info):
        return self.ast.SeqIndex(s, ind, position, info, self.NoTrafos)

    def SeqTake(self, s, end, position, info):
        return self.ast.SeqTake(s, end, position, info, self.NoTrafos)

    def SeqDrop(self, s, end, position, info):
        return self.ast.SeqDrop(s, end, position, info, self.NoTrafos)

    def Add(self, left, right, position, info):
        return self.ast.Add(left, right, position, info, self.NoTrafos)

    def Sub(self, left, right, position, info):
        return self.ast.Sub(left, right, position, info, self.NoTrafos)

    def Mul(self, left, right, position, info):
        return self.ast.Mul(left, right, position, info, self.NoTrafos)

    def Div(self, left, right, position, info):
        return self.ast.Div(left, right, position, info, self.NoTrafos)

    def Mod(self, left, right, position, info):
        return self.ast.Mod(left, right, position, info, self.NoTrafos)

    def And(self, left, right, position, info):
        return self.ast.And(left, right, position, info, self.NoTrafos)

    def Or(self, left, right, position, info):
        return self.ast.Or(left, right, position, info, self.NoTrafos)

    def If(self, cond, thn, els, position, info):
        thn_seqn = self.Seqn([thn], position, info)
        els_seqn = self.Seqn([els], position, info)
        return self.ast.If(cond, thn_seqn, els_seqn, position, info, self.NoTrafos)

    def TrueLit(self, position, info):
        return self.ast.TrueLit(position, info, self.NoTrafos)

    def FalseLit(self, position, info):
        return self.ast.FalseLit(position, info, self.NoTrafos)

    def NullLit(self, position, info):
        return self.ast.NullLit(position, info, self.NoTrafos)

    def Forall(self, variables, triggers, exp, position, info):
        res = self.ast.Forall(self.to_seq(variables), self.to_seq(triggers),
                               exp, position, info, self.NoTrafos)
        if res.isPure():
            return res
        else:
            desugared = self.to_list(self.QPs.desugarSourceQuantifiedPermissionSyntax(res))
            result = self.TrueLit(position, info)
            for qp in desugared:
                result = self.And(result, qp, position, info)
            return result

    def Exists(self, variables, exp, position, info):
        res = self.ast.Exists(self.to_seq(variables), exp, position, info, self.NoTrafos)
        return res

    def Trigger(self, exps, position, info):
        return self.ast.Trigger(self.to_seq(exps), position, info, self.NoTrafos)

    def While(self, cond, invariants, locals, body, position, info):
        body_with_locals = self.Seqn([body], position, info, locals)
        return self.ast.While(cond, self.to_seq(invariants),
                              body_with_locals, position, info, self.NoTrafos)

    def Let(self, variable, exp, body, position, info):
        return self.ast.Let(variable, exp, body, position, info, self.NoTrafos)

    def from_option(self, option):
        if option == self.none:
            return None
        else:
            return option.get()

    def to_function0(self, func):
        func0 = Function0()
        func0.apply = types.MethodType(func, func0)
        result = self.jvm.get_proxy('scala.Function0', func0)
        return result

    def SimpleInfo(self, comments):
        return self.ast.SimpleInfo(self.to_seq(comments))

    def ConsInfo(self, head, tail):
        return self.ast.ConsInfo(head, tail)

    def to_position(self, expr, vias, error_string: str=None,
                    rules: Rules=None, file: str = None):
        if expr is None:
            return self.NoPosition
        if not hasattr(expr, 'lineno'):
            # TODO: this should never happen (because all nodes that the parser
            # creates have this field), but it does, because in the SIF code we
            # create artificial ast.Name objects which don't have it. That
            # should probably be changed.
            return self.NoPosition
        if not file:
            file = str(self.sourcefile)
        path = self.java.nio.file.Paths.get(file, [])
        start = self.ast.LineColumnPosition(expr.lineno, expr.col_offset)
        id = error_manager.add_error_information(
            expr, list(vias), error_string, rules)
        if hasattr(expr, 'end_lineno') and hasattr(expr, 'end_col_offset'):
            end = self.ast.LineColumnPosition(expr.end_lineno,
                                              expr.end_col_offset)
            end = self.scala.Some(end)
        else:
            end = self.none
        return self.ast.IdentifierPosition(path, start, end, id)

    def is_heap_dependent(self, expr) -> bool:
        """
        Checks if the given expression contains an access to a heap location.
        Does NOT check for calls to heap-dependent functions.
        """
        for n in [expr] + self.to_list(expr.subnodes()):
            if isinstance(n, self.ast.LocationAccess):
                return True
        return False




"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Obligation translator in methods."""


import ast

from typing import List

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonMethod,
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Method,
    Stmt,
    Position,
    VarDecl,
)
from nagini_translation.translators.obligation.common import (
    CommonObligationTranslator,
)
from nagini_translation.translators.obligation.fork import (
    ObligationMethodForkConstructor,
)
from nagini_translation.translators.obligation.method_node import (
    ObligationMethod,
    ObligationMethodNodeConstructor,
)
from nagini_translation.translators.obligation.method_call_node import (
    ObligationMethodCall,
    ObligationMethodCallNodeConstructor,
)
from nagini_translation.translators.obligation.types.base import (
    ObligationInstance,
)
from nagini_translation.translators.obligation.obligation_info import (
    BaseObligationInfo,
)
from nagini_translation.translators.obligation.utils import (
    find_method_by_sil_name,
)


class MethodObligationTranslator(CommonObligationTranslator):
    """Class for translating obligations in methods."""

    def _get_obligation_info(self, ctx: Context) -> BaseObligationInfo:
        return ctx.actual_function.obligation_info

    def _create_obligation_instance_use(
            self, obligation_instance: ObligationInstance,
            ctx: Context) -> sil.InhaleExhale:
        return obligation_instance.get_use_method(ctx)

    def create_method_node(     # pylint: disable=too-many-arguments,too-many-locals
            self, ctx: Context, name: str,
            original_args: List[VarDecl], returns: List[VarDecl],
            pres: List[Expr], posts: List[Expr],
            local_vars: List[VarDecl], body: List[Stmt],
            position: Position, info: Info,
            method: PythonMethod = None,
            overriding_check: bool = False) -> Method:
        """Construct method AST node with additional obligation stuff."""
        if method is None:
            method = find_method_by_sil_name(ctx, name)
        if method is None:
            # Assume that this is a method that is never called from
            # Python and, as a result, does not need obligation stuff.
            return self.viper.Method(
                name, original_args, returns, pres, posts, local_vars, body,
                position, info)

        assert (ctx.current_function is None or
                ctx.current_function == method)
        old_method = ctx.current_function
        ctx.current_function = method

        obligation_method = ObligationMethod(
            name, original_args, returns, pres, posts, local_vars, body)
        constructor = ObligationMethodNodeConstructor(
            obligation_method, method, self, ctx, self._obligation_manager,
            position, info, overriding_check)
        constructor.add_obligations()
        node = constructor.construct_node()

        ctx.current_function = old_method
        return node

    def create_method_call_node(
            self, ctx: Context, method_name: str, original_args: List[Expr],
            targets: List[Expr], position: Position, info: Info,
            target_method: PythonMethod = None,
            target_node: ast.Call = None) -> List[Stmt]:
        """Construct a method call AST node with obligation stuff."""
        obligation_method_call = ObligationMethodCall(
            method_name, original_args, targets)
        constructor = ObligationMethodCallNodeConstructor(
            obligation_method_call, position, info, self, ctx,
            self._obligation_manager, target_method, target_node)
        constructor.construct_call()
        return constructor.get_statements()

    def create_method_fork(self, ctx: Context, targets, thread: Expr,
                           position: Position, info: Info,
                           target_node: ast.Call=None) -> List[Stmt]:
        constructor = ObligationMethodForkConstructor(targets, thread, position, info,
                                                      self, ctx, self._obligation_manager,
                                                      target_node)
        constructor.construct_fork()
        return constructor.get_statements()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Functions and classes needed for writing obligation related contracts."""

# pragma pylint: disable=invalid-name,unused-argument


from typing import Union
from nagini_contracts.thread import Thread


OBLIGATION_CONTRACT_FUNCS = [
    'MustTerminate',
    'MustRelease',
    'Level',
    'WaitLevel',
]


class BaseLock:
    """A base class for locks."""


class LevelType:
    """A type returned by ``Level`` function."""

    def __lt__(self, other: 'LevelType') -> bool:
        """We allow to compare only ``LevelType`` objects."""


def WaitLevel() -> LevelType:
    """The wait level of the current thread."""


def Level(l: Union[BaseLock, Thread]) -> LevelType:
    """Level of the given lock or thread."""


def MustRelease(lock: BaseLock, measure: int = None) -> bool:
    """An obligation to release a ``lock`` in ``measure`` steps."""


def MustTerminate(measure: int) -> bool:
    """An obligation to terminate in ``measure`` steps."""


__all__ = (
    'MustRelease',
    'MustTerminate',
    'LevelType',
    'WaitLevel',
    'Level',
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_translation.analyzer import PythonModule, PythonVar
from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.program_nodes import (
    ContainerInterface,
    PythonMethod,
    PythonNode,
)
from nagini_translation.lib.resolver import get_target
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.abstract import (
    Context,
    Expr,
    TranslatorConfig,
)
from nagini_translation.translators.call import CallTranslator
from nagini_translation.translators.contract import ContractTranslator
from nagini_translation.translators.expression import ExpressionTranslator
from nagini_translation.translators.io_operation import (
    IOOperationTranslator,
)
from nagini_translation.translators.method import MethodTranslator
from nagini_translation.translators.obligation import (
    ObligationTranslator,
)
from nagini_translation.translators.permission import PermTranslator
from nagini_translation.translators.predicate import PredicateTranslator
from nagini_translation.translators.program import ProgramTranslator
from nagini_translation.translators.pure import PureTranslator
from nagini_translation.translators.statement import StatementTranslator
from nagini_translation.translators.type import TypeTranslator
from nagini_translation.translators.type_domain_factory import (
    TypeDomainFactory,
)
from typing import List, Optional, Set


class Translator:
    """
    Translates a Python AST to a Silver AST.
    This class serves as the public interface of the entire translator.
    The functionality is implemented in several specialized translators; this
    class only sets up the inner translator structure and forwards calls from
    the public interface to the responsible specialized translators.
    """

    def __init__(self, jvm: JVM, source_file: str, type_info: TypeInfo,
                 viper_ast: ViperAST):
        config = TranslatorConfig(self)
        config.pure_translator = PureTranslator(config, jvm, source_file,
                                                type_info, viper_ast)
        config.call_translator = CallTranslator(config, jvm, source_file,
                                                type_info, viper_ast)
        config.contract_translator = ContractTranslator(config, jvm,
                                                        source_file,
                                                        type_info, viper_ast)
        config.expr_translator = ExpressionTranslator(config, jvm, source_file,
                                                      type_info, viper_ast)
        config.pred_translator = PredicateTranslator(config, jvm, source_file,
                                                     type_info, viper_ast)
        config.io_operation_translator = IOOperationTranslator(
            config, jvm, source_file, type_info, viper_ast)
        config.obligation_translator = ObligationTranslator(
            config, jvm, source_file, type_info, viper_ast)
        config.stmt_translator = StatementTranslator(config, jvm, source_file,
                                                     type_info, viper_ast)
        config.perm_translator = PermTranslator(config, jvm, source_file,
                                                type_info, viper_ast)
        config.type_translator = TypeTranslator(config, jvm, source_file,
                                                type_info, viper_ast)
        config.prog_translator = ProgramTranslator(config, jvm, source_file,
                                                   type_info, viper_ast)
        config.method_translator = MethodTranslator(config, jvm, source_file,
                                                    type_info, viper_ast)
        config.type_factory = TypeDomainFactory(viper_ast, self)
        self.obligation_translator = config.obligation_translator
        self.prog_translator = config.prog_translator
        self.expr_translator = config.expr_translator

    def translate_program(self, modules: List[PythonModule], sil_progs: List,
                          selected: Set[str] = None,
                          ignore_global: bool = False,
                          arp: bool = False) -> 'silver.ast.Program':
        ctx = Context()
        ctx.current_class = None
        ctx.current_function = None
        ctx.module = modules[0]
        ctx.arp = arp
        return self.prog_translator.translate_program(modules, sil_progs, ctx,
                                                      selected, ignore_global)

    def translate_pythonvar_decl(self, var: PythonVar,
            module: PythonModule) -> 'silver.ast.LocalVarDecl':
        # We need a context object here
        ctx = Context()
        ctx.module = module
        return self.expr_translator.translate_pythonvar_decl(var, ctx)

    def translate_pythonvar_ref(self, var: PythonVar,
                                module: PythonModule,
                                node: ast.AST, ctx: Context) -> Expr:
        # We need a context object here
        if not ctx:
            ctx = Context()
            ctx.module = module
        return self.expr_translator.translate_pythonvar_ref(var, node, ctx)

    def to_position(self, node: ast.AST, ctx: Context) -> 'silver.ast.Position':
        return self.expr_translator.to_position(node, ctx)

    def no_position(self, ctx: Context) -> 'silver.ast.Position':
        return self.to_position(None, ctx)

    def to_info(self, comments: List[str], ctx: Context) -> 'silver.ast.Info':
        return self.expr_translator.to_info(comments, ctx)

    def no_info(self, ctx: Context) -> 'silver.ast.Info':
        return self.to_info([], ctx)

    def set_required_names(self, name: str, required_names: Set[str]) -> None:
        """
        Registers that the native Silver method/function named 'name' depends
        on the methods/functions in the given set.
        """
        self.prog_translator.required_names[name] = required_names

    def create_obligation_info(self, method: PythonMethod) -> object:
        """
        Create an obligation info for method. This method should be
        called during the processing stage of the method before any
        translation is done.

        This return type of this method is ``object`` to indicate that
        the returned value is opaque for all code except obligation
        translator.
        """
        return self.obligation_translator.create_obligation_info(method)

    def get_target(self, node: ast.AST,
                   containers: List[ContainerInterface],
                   container: PythonNode) -> PythonNode:
        return get_target(node, containers, container)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

from nagini_translation.lib.typedefs import Stmt
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.program import ProgramTranslator

from typing import List

class SIFProgramTranslator(ProgramTranslator):
    """
    Extended AST version of program translator.
    """

    def _create_inherit_check_postamble(self, stmts: List[Stmt],
                                        end_lbl: 'silver.ast.Label',
                                        ctx: Context) -> None:
        pass



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""A helper class for translating IO operation Open."""


import ast

from typing import cast, List

from nagini_translation.lib.context import Context
from nagini_translation.lib.io_context import IOOpenContext
from nagini_translation.lib.program_nodes import (
    PythonIOOperation,
    PythonVar,
)
from nagini_translation.lib.resolver import get_target
from nagini_translation.lib.typedefs import (
    Expr,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.io_operation.common import (
    IOOperationCommonTranslator,
)
from nagini_translation.translators.io_operation.utils import (
    raise_invalid_operation_use,
)


def _get_opened_operation(
        node: ast.Call, ctx: Context) -> PythonIOOperation:
    """Get the operation that is being opened."""
    if (len(node.args) == 1 and
            isinstance(node.args[0], ast.Call) and
            isinstance(node.args[0].func, ast.Name)):
        containers = [ctx.module]
        containers.extend(ctx.module.get_included_modules())
        target = get_target(node.args[0], containers, None)
        if isinstance(target, PythonIOOperation):
            return target
    raise_invalid_operation_use('open_non_io_operation', node)


class Opener:
    """A class responsible for translating ``Open`` call."""

    def __init__(self, node: ast.Call, ctx: Context,
                 translator: IOOperationCommonTranslator,
                 full_perm: 'viper_ast.FullPerm') -> None:
        self._node = node
        self._ctx = ctx
        self._operation = _get_opened_operation(node, ctx)
        self._operation_call = cast(ast.Call, node.args[0])
        self._translator = translator
        self._full_perm = full_perm

    @property
    def _io_ctx(self) -> IOOpenContext:
        return self._ctx.io_open_context

    @property
    def _viper(self) -> ViperAST:
        return self._translator.viper

    @property
    def _position(self) -> 'viper_ast.Position':
        return self._translator.to_position(self._node, self._ctx)

    @property
    def _info(self) -> 'viper_ast.Info':
        return self._translator.no_info(self._ctx)

    def translate(self) -> StmtsAndExpr:
        """Translate IO operation open."""
        self._io_ctx.start_io_operation_open()

        statements = []

        self._check()
        sil_args = self._translate_arguments()
        statements.append(self._translate_exhale(sil_args))
        statements.extend(self._translate_inhale(sil_args))

        self._io_ctx.stop_io_operation_open()
        return (statements, None)

    def _check(self) -> None:
        """Check that operation open is well-formed."""
        if self._operation.is_basic():
            raise_invalid_operation_use('open_basic_io_operation',
                                        self._node)
        if (len(self._operation_call.args) !=
                len(self._operation.get_parameters())):
            raise_invalid_operation_use('result_used_argument', self._node)

    def _translate_arguments(self) -> List[Expr]:
        return self._translator.translate_args(
            self._operation_call.args, self._operation.get_parameters(),
            self._ctx)

    def _translate_exhale(self, sil_args: List[Expr]) -> Stmt:
        """Translate exhale of IO operation."""
        predicate = self._translator.create_predicate_access(
            self._operation.sil_name, sil_args, self._full_perm,
            self._node, self._ctx)
        exhale = self._viper.Exhale(predicate, self._position, self._info)
        return exhale

    def _translate_inhale(self, sil_args: List[Expr]) -> List[Stmt]:
        """Translate inhale of IO operation."""
        with self._ctx.aliases_context():
            self._define_input_aliases()
            self._define_output_aliases(sil_args)
            self._define_existential_variables()
            self._ctx.inlined_calls.append(self._operation)
            body = self._translate_body()
            self._ctx.inlined_calls.pop()
        statements = self._emit_existential_variable_definitions()
        statements.append(self._emit_body_inhale(body))
        return statements

    def _define_existential_variables(self) -> None:
        """Define fresh local variables for stuff mentioned in IOExists.

        Make sure that they have fresh silver names. Add them to the
        context and variable aliases.
        """
        io_existential_vars = dict(
            (creator.name, creator.create_variable_instance())
            for creator in self._operation.get_io_existentials()
        )
        for name, var in io_existential_vars.items():
            self._add_local_var(name, var)

    def _define_input_aliases(self) -> None:
        """Set up aliases for input.

        Here we use existential variables during translation that are
        later replaced by silver expressions that were provided as
        arguments.
        """
        self._translator.set_up_io_operation_input_aliases(
            self._operation, self._operation_call, self._ctx)

    def _define_output_aliases(self, sil_args: List[Expr]) -> None:
        """Set up aliases for output.

        The same idea as with IOExists stuff, just we immediately
        provide their definitions because we know them.
        """
        for result in self._operation.get_results():
            name = result.name
            var = PythonVar(name, result.node, result.type)
            self._add_local_var(name, var)
            getter = self._translator.create_result_getter(
                self._operation_call, result, self._ctx,
                sil_args=sil_args)
            self._io_ctx.define_variable(name, getter)

    def _translate_body(self) -> Expr:
        """Translate body.

        During translation defining getters are stored in the context
        and variables are replaced by variables created earlier. Note
        that existentials defined by ``existential == expression`` are
        not allowed because both use cases (fields and Result()) are
        forbidden inside IO operations.
        """
        statements, body_expression = self._translator.translate_expr(
            self._operation.get_body(), self._ctx,
            target_type=self._translator.viper.Bool)
        assert not statements
        return body_expression

    def _emit_existential_variable_definitions(self) -> List[Stmt]:
        """Emit existential variable definitions.

        Emit equalities between created variables and their
        corresponding defining getters.
        """
        statements = []
        for var, definition in self._io_ctx.get_ordered_variable_defs():
            assignment = self._viper.LocalVarAssign(
                var.ref(), definition, self._position, self._info)
            statements.append(assignment)
        return statements

    def _emit_body_inhale(self, body: Expr) -> Stmt:
        """Emit inhale of the translate IO operation body."""
        inhale = self._viper.Inhale(
            body, self._position, self._info)
        return inhale

    def _add_local_var(self, name, var) -> None:
        sil_name = self._ctx.actual_function.get_fresh_name(name)
        var.process(sil_name, self._translator.translator)
        self._ctx.actual_function.locals[sil_name] = var
        self._io_ctx.add_variable(name, var)
        self._ctx.set_alias(name, var)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Common code for constructing Silver nodes with obligation stuff."""


import ast

from typing import List, Union

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.config import obligation_config
from nagini_translation.lib.context import Context
from nagini_translation.lib.errors import Rules
from nagini_translation.lib.program_nodes import (
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Info,
    Stmt,
    Position,
)
from nagini_translation.lib.util import (
    pprint,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)
from nagini_translation.translators.obligation.measures import (
    MeasureMap,
)
from nagini_translation.translators.obligation.types import (
    must_terminate,
)
from nagini_translation.translators.obligation.obligation_info import (
    PythonMethodObligationInfo,
)


class StatementNodeConstructorBase:
    """Common functionality for loop and method call constructors."""

    def __init__(
            self, translator: 'AbstractTranslator', ctx: Context,
            obligation_manager: ObligationManager,
            position: Position, info: Info,
            default_position_node: ast.AST) -> None:
        self._position = position
        """Default position."""
        self._info = info
        """Default info."""
        self._translator = translator
        self._ctx = ctx
        self._obligation_manager = obligation_manager
        self._default_position_node = default_position_node
        self._statements = []

    def get_statements(self) -> List[Stmt]:
        """Get all generated statements."""
        return self._statements

    def _save_must_terminate_amount(
            self, amount_var: PythonVar) -> None:
        """Save the current permission amount to a variable."""
        if obligation_config.disable_termination_check:
            return
        predicate = self._get_must_terminate_predicate()
        assign = sil.Assign(amount_var, sil.CurrentPerm(predicate))
        info = self._to_info('Save current MustTerminate amount.')
        self._append_statement(assign, info=info)

    def _reset_must_terminate(self, amount_var: PythonVar) -> None:
        """Reset ``MustTerminate`` permission to its original level.

        .. note::

            Implication is needed because in Silicon if callee took all
            permission, the ``exhale acc(..., none)`` would fail, even
            though this exhale does nothing.
        """
        if obligation_config.disable_termination_check:
            return
        predicate = self._get_must_terminate_predicate()
        original_amount = sil.PermVar(amount_var)
        perm = sil.CurrentPerm(predicate) - original_amount
        exhale = sil.Exhale(sil.Implies(
            sil.CurrentPerm(predicate) > sil.NoPerm(),
            sil.Acc(predicate, perm)))
        info = self._to_info('Reset MustTerminate amount to original level.')
        self._append_statement(exhale, info=info)

    @property
    def _viper(self) -> ViperAST:
        return self._translator.viper

    @property
    def _must_terminate(self) -> must_terminate.MustTerminateObligation:
        return self._obligation_manager.must_terminate_obligation

    @property
    def _obligation_info(self) -> PythonMethodObligationInfo:
        """Get the surrounding method obligation info."""
        return self._ctx.current_function.obligation_info

    @property
    def _method_measure_map(self) -> MeasureMap:
        """Get the surrounding method measure map."""
        return self._obligation_info.method_measure_map

    def _get_must_terminate_predicate(self) -> sil.PredicateAccess:
        cthread = self._obligation_info.current_thread_var
        return self._must_terminate.create_predicate_access(cthread)

    def _to_position(
            self, node: ast.AST = None,
            conversion_rules: Rules = None,
            error_node: Union[str, ast.AST] = None) -> Position:
        error_string = None
        if error_node is not None:
            if isinstance(error_node, ast.AST):
                error_string = pprint(error_node)
            else:
                error_string = error_node
        return self._translator.to_position(
            node or self._default_position_node, self._ctx,
            error_string=error_string, rules=conversion_rules)

    def _to_info(self, template, *args, **kwargs) -> Info:
        return self._translator.to_info(
            [template.format(*args, **kwargs)], self._ctx)

    def _append_statement(
            self, statement: sil.Statement,
            position: Position = None, info: Info = None) -> None:
        translated = statement.translate(
            self._translator, self._ctx,
            position or self._position,
            info or self._info)
        self._statements.append(translated)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Built-in IO operations."""

# pragma pylint: disable=invalid-name,unused-argument


from typing import Callable, Tuple, TypeVar

from nagini_contracts.contracts import (
    ContractOnly,
    Ensures,
    Ghost,
    Result,
    Requires,
)
from nagini_contracts.io_contracts import (
    ctoken,
    IOExists1,
    IOExists2,
    IOExists3,
    IOOperation,
    Place,
    Terminates,
    token,
)
from nagini_contracts.obligations import (
    MustTerminate,
)


@IOOperation
def no_op_io(
        t_pre: Place,
        t_post: Place = Result()) -> bool:
    """No op IO operation."""
    Terminates(True)


@Ghost
@ContractOnly
def NoOp(t_pre: Place) -> Place:
    """Perform ``no_op_io``."""
    IOExists1(Place)(
        lambda t_post: (
            Requires(
                token(t_pre, 1) and
                no_op_io(t_pre, t_post) and
                MustTerminate(1)
            ),
            Ensures(
                token(t_post) and
                t_post == Result()
            ),
        )
    )


@IOOperation
def split_io(
        t_pre: Place,
        t_post1: Place = Result(),
        t_post2: Place = Result()) -> bool:
    """Split one place into two places."""
    Terminates(True)


@Ghost
@ContractOnly
def Split(t_pre: Place) -> Tuple[Place, Place]:
    """Perform ``split_io``."""
    IOExists2(Place, Place)(
        lambda t_post1, t_post2: (
            Requires(
                token(t_pre, 1) and
                split_io(t_pre, t_post1, t_post2) and
                MustTerminate(1)
            ),
            Ensures(
                token(t_post1) and
                t_post1 == Result()[0] and
                token(t_post2) and
                t_post2 == Result()[1]
            ),
        )
    )


@IOOperation
def join_io(
        t_pre1: Place,
        t_pre2: Place,
        t_post: Place = Result()) -> bool:
    """Join to input places into one output place."""
    Terminates(True)


@Ghost
@ContractOnly
def Join(t_pre1: Place, t_pre2: Place) -> Place:
    """Perform ``join_io``."""
    IOExists1(Place)(
        lambda t_post: (
            Requires(
                token(t_pre1, 1) and
                token(t_pre2, 1) and
                join_io(t_pre1, t_pre2, t_post) and
                MustTerminate(1)
            ),
            Ensures(
                token(t_post) and
                t_post == Result()
            ),
        )
    )


@IOOperation
def gap_io(
        t_pre: Place,
        t_post: Place = Result()) -> bool:
    """Convert ``token`` to ``ctoken``."""
    Terminates(True)


@Ghost
@ContractOnly
def Gap(t_pre: Place) -> Place:
    """Perform ``gap_io``."""
    IOExists1(Place)(
        lambda t_post: (
            Requires(
                token(t_pre, 1) and
                gap_io(t_pre, t_post) and
                MustTerminate(1)
            ),
            Ensures(
                ctoken(t_post) and
                t_post == Result()
            ),
        )
    )


@IOOperation
def end_io(t_pre: Place) -> bool:
    """Terminate Petri Net."""
    Terminates(True)


@Ghost
@ContractOnly
def End(t_pre: Place) -> Place:
    """Perform ``end_io``."""
    Requires(
        token(t_pre, 1) and
        end_io(t_pre) and
        MustTerminate(1)
    )


@IOOperation
def set_var_io(
        t_pre: Place,
        value: int,
        result: int = Result(),
        t_post: Place = Result()) -> bool:
    """Set variable given at ``result`` position to provided ``value``."""
    Terminates(True)


@Ghost
@ContractOnly
def SetVar(t_pre: Place, value: int) -> Tuple[int, Place]:
    """Perform ``set_var_io``.

    .. note::

        Mypy does not allow generics as function arguments. Therefore,
        we have to use a concrete type for ``value``.
    """
    IOExists2(Place, int)(
        lambda t_post, result: (
            Requires(
                token(t_pre, 1) and
                set_var_io(t_pre, value, result, t_post) and
                MustTerminate(1)
            ),
            Ensures(
                token(t_post) and
                t_post == Result()[1] and
                result == Result()[0] and
                value == result
            ),
        )
    )


T = TypeVar('T')
V = TypeVar('V')


@IOOperation
def eval_io(
        t_pre: Place,
        func: Callable[[T], V],
        arg: T,
        result: object = Result(),
        t_post: Place = Result()) -> bool:
    """Evaluate func in the current state."""
    Terminates(True)


@Ghost
@ContractOnly
def Eval(t_pre: Place, func: Callable[[T], V], arg: T) -> Tuple[V, Place]:
    IOExists2(Place, object)(
        lambda t_post, result: (
            Requires(
                token(t_pre, 1) and
                eval_io(t_pre, func, arg, result, t_post) and
                MustTerminate(1)
            ),
            Ensures(
                token(t_post) and
                t_post is Result()[1] and
                result is Result()[0]
                # and resul result == func(arg)
                # This is part of the contract but guaranteed via additional inhales after
                # any call to Eval, and cannot be part of the contract written here
                # because the fact that func can have an arbitrary precondition would make
                # the contract of Eval not well-formed.
            ),
        )
    )


"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Obligation translator in loops."""


import ast

from typing import List, Union

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Stmt,
    VarDecl,
)
from nagini_translation.translators.obligation.common import (
    CommonObligationTranslator
)
from nagini_translation.translators.obligation.loop_node import (
    ObligationLoop,
    ObligationLoopNodeConstructor,
)
from nagini_translation.translators.obligation.types.base import (
    ObligationInstance,
)
from nagini_translation.translators.obligation.obligation_info import (
    BaseObligationInfo,
    PythonLoopObligationInfo,
)


class LoopObligationTranslator(CommonObligationTranslator):
    """Class for translating obligations in loops."""

    def enter_loop_translation(
            self, node: Union[ast.While, ast.For], ctx: Context,
            err_var: PythonVar = None) -> None:
        """Update context with info needed to translate loop."""
        info = PythonLoopObligationInfo(
            self._obligation_manager, node, self, ctx.actual_function,
            err_var)
        info.traverse_invariants()
        ctx.obligation_context.push_loop_info(info)

    def leave_loop_translation(self, ctx: Context) -> None:
        """Remove loop translation info from context."""
        ctx.obligation_context.pop_loop_info()

    def _get_obligation_info(self, ctx: Context) -> BaseObligationInfo:
        return ctx.obligation_context.current_loop_info

    def _create_obligation_instance_use(
            self, obligation_instance: ObligationInstance,
            ctx: Context) -> sil.InhaleExhale:
        return obligation_instance.get_use_loop(ctx)

    def create_while_node(
            self, ctx: Context, cond: Expr,
            invariants: List[Expr],
            local_vars: List[VarDecl],
            body: List[Stmt], node: Union[ast.While, ast.For]) -> List[Stmt]:
        """Construct a while loop AST node with obligation stuff."""
        obligation_loop = ObligationLoop(cond, invariants, local_vars, body)
        constructor = ObligationLoopNodeConstructor(
            obligation_loop, node, self, ctx, self._obligation_manager)
        constructor.construct_loop()
        return constructor.get_statements()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_contracts.contracts import GHOST_PREFIX


contract_keywords = frozenset(["Requires", "Ensures", "Exsures",
                               "Invariant",
                               "Assume", "Assert",
                               "Old",
                               "Result",
                               "Implies",
                               "Forall", "Exists",
                               "Acc",
                               "Fold", "Unfold", "Unfolding",
                               "Pure", "Predicate",
                               "Ghost",])


class GhostCollector(ast.NodeVisitor):
    """
    AST visitor to collect ghost variables and functions.
    """
    def __init__(self):
        self.ghost_vars = set()
        self.ghost_funcs = set()

    @classmethod
    def collect(cls, tree):
        """
        Collects and returns two sets containing the names of ghost variables
        and functions.
        :param tree: The AST of the program.
        :return: Tuple of sets (ghost_vars, ghost_funcs).
        """
        collector = cls()
        collector.visit(tree)
        return collector.ghost_vars, collector.ghost_funcs

    def visit_Name(self, node):
        if node.id.startswith(GHOST_PREFIX):
            self.ghost_vars.add(node.id)

    def visit_Attribute(self, node):
        if node.attr.startswith(GHOST_PREFIX):
            self.ghost_vars.add(node.attr)

    def visit_FunctionDef(self, node):
        decorator_list = [dec.id for dec in node.decorator_list]
        if "Ghost" in decorator_list:
            self.ghost_funcs.add(node.name)


class NoOpCollector(ast.NodeVisitor):
    """
    AST visitor that collects @pure, @predicate and @ghost annotations.
    """
    def __init__(self, ghost_vars, ghost_funcs):
        self.ghost_vars = ghost_vars
        self.ghost_funcs = ghost_funcs
        self.noop_stmts = set()
        self._cur_stmt = None

    @property
    def _cur_stmt(self):
        return self.__cur_stmt

    @_cur_stmt.setter
    def _cur_stmt(self, stmt):
        assert stmt is None or \
               isinstance(stmt, ast.stmt), "_cur_stmt is not a stmt or None"
        self.__cur_stmt = stmt

    @classmethod
    def collect(cls, tree, ghost_vars, ghost_funcs):
        """
        Collects and returns statements that need to be turned into no-ops
        (i.e., deleted from the AST).

        :param tree: The AST of the program.
        :param ghost_vars: A set containing the names of ghost variables.
        :param ghost_funcs: A set containing the names of ghost functions.
        :return: A set containing ast.stmt's that need to be deleted.
        """
        collector = cls(ghost_vars, ghost_funcs)
        collector.visit(tree)
        return collector.noop_stmts

    def visit_Module(self, node):
        for stmt in node.body:
            self._cur_stmt = stmt
            self.visit(stmt)

    def visit_Name(self, node):
        if node.id.startswith(GHOST_PREFIX):
            self.noop_stmts.add(self._cur_stmt)

    def visit_Attribute(self, node):
        if node.attr.startswith(GHOST_PREFIX):
            self.noop_stmts.add(self._cur_stmt)

    def visit_FunctionDef(self, node):
        if node.name in self.ghost_funcs:
            self.noop_stmts.add(node)
            return

        self.generic_visit(node)

    def visit_Call(self, node):
        func_name = None
        if isinstance(node.func, ast.Name):
            func_name = node.func.id
        elif isinstance(node.func, ast.Attribute):
            func_name = node.func.attr

        if func_name in contract_keywords or func_name in self.ghost_funcs:
            self.noop_stmts.add(self._cur_stmt)
            return

        self.generic_visit(node)

    def visit_Delete(self, node):
        self._cur_stmt = node
        self.generic_visit(node)

    def visit_Assign(self, node):
        self._cur_stmt = node
        self.generic_visit(node)

    def visit_AugAssign(self, node):
        self._cur_stmt = node
        self.generic_visit(node)

    def visit_Expr(self, node):
        self._cur_stmt = node
        self.generic_visit(node)

    def visit_For(self, node):
        self._cur_stmt = node
        self.generic_visit(node)

    def visit_While(self, node):
        self._cur_stmt = node
        self.generic_visit(node)

    def visit_If(self, node):
        self._cur_stmt = node
        self.generic_visit(node)


class EmptyBodyCollector(ast.NodeVisitor):
    """
    Collects statement that have an empty body due to AST rewriting.
    """
    def __init__(self):
        self.noop_stmts = set()

    @classmethod
    def collect(cls, tree):
        """
        Collects and returns statements that need to be removed, because their
        body does not contain any statements (e.g., after removing all
        statements involving ghost variables).

        :param tree: The AST of the program.
        :return: Set containing ast.stmt's that need to be deleted.
        """
        collector = cls()
        collector.visit(tree)
        return collector.noop_stmts

    def visit_FunctionDef(self, node):
        if not node.body:
            self.noop_stmts.add(node)
            return

        self.generic_visit(node)

    def visit_ClassDef(self, node):
        if not node.body:
            self.noop_stmts.add(node)
            return

        self.generic_visit(node)

    def visit_For(self, node):
        if not node.body:
            self.noop_stmts.add(node)
            return

        self.generic_visit(node)

    def visit_While(self, node):
        if not node.body:
            self.noop_stmts.add(node)
            return

        self.generic_visit(node)

    def visit_If(self, node):
        if not node.body:
            self.noop_stmts.add(node)
            return

        self.generic_visit(node)

    def visit_Try(self, node):
        if not node.body:
            self.noop_stmts.add(node)
            return

        self.generic_visit(node)


class StmtEraser(ast.NodeTransformer):
    """
    AST visitor that deletes statements passed by 'noop_stmts'.
    """
    def __init__(self, noop_stmts):
        self.noop_stmts = noop_stmts

    @classmethod
    def erase(cls, tree, noop_stmts):
        """
        Removes statements in 'noop_stmts' from the AST.

        :param tree: The AST to be modified.
        :param noop_stmts: The statements to be removed.
        :return: The modified AST.
        """
        eraser = cls(noop_stmts)
        return eraser.visit(tree)

    def visit(self, node):
        if isinstance(node, ast.stmt) and node in self.noop_stmts:
            return None

        self.generic_visit(node)
        return node


def transform_ast(tree):
    """
    Applies necessary transformations to the AST to make code containing
    contracts run correctly.
    :param tree: source node of the AST
    :return: source node of the transformed AST
    """
    transformed_ast = tree
    # Collect ghost variables and functions.
    ghost_vars, ghost_funcs = GhostCollector.collect(tree)
    # Collect no-op statements.
    noop_stmts = NoOpCollector.collect(transformed_ast, ghost_vars, ghost_funcs)
    while noop_stmts:
        # Remove no-op statements from the AST.
        transformed_ast = StmtEraser.erase(transformed_ast, noop_stmts)
        # Collect statements with an empty body.
        noop_stmts = EmptyBodyCollector.collect(transformed_ast)

    return transformed_ast



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Functions and classes needed for writing IO related contracts."""

# pragma pylint: disable=too-many-arguments,invalid-name,unused-argument,too-many-locals


from typing import Any, TypeVar, Generic, Callable, Type


BUILTIN_IO_OPERATIONS = (
    'no_op_io',
    'split_io',
    'join_io',
    'gap_io',
    'set_var_io',
)


IO_OPERATION_PROPERTY_FUNCS = [
    'Terminates',
    'TerminationMeasure',
]


IO_CONTRACT_FUNCS = [
    'ctoken',
    'token',
    'Open',
    'Eval',
    'eval_io',
]


T = TypeVar('T')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')
T5 = TypeVar('T5')
T6 = TypeVar('T6')
T7 = TypeVar('T7')
T8 = TypeVar('T8')
T9 = TypeVar('T9')
T10 = TypeVar('T10')
T11 = TypeVar('T11')
T12 = TypeVar('T12')
T13 = TypeVar('T13')
T14 = TypeVar('T14')
T15 = TypeVar('T15')


class IOExists1(Generic[T1]):
    """``IOExists`` for defining 1 IO existential variable."""

    def __init__(
            self,
            t1: Type[T1]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1], Any]) -> bool:
        pass


class IOExists2(Generic[T1, T2]):
    """``IOExists`` for defining 2 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1, T2], Any]) -> bool:
        pass


class IOExists3(Generic[T1, T2, T3]):
    """``IOExists`` for defining 3 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1, T2, T3], Any]) -> bool:
        pass


class IOExists4(Generic[T1, T2, T3, T4]):
    """``IOExists`` for defining 4 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1, T2, T3, T4], Any]) -> bool:
        pass


class IOExists5(Generic[T1, T2, T3, T4, T5]):
    """``IOExists`` for defining 5 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1, T2, T3, T4, T5], Any]) -> bool:
        pass


class IOExists6(Generic[T1, T2, T3, T4, T5, T6]):
    """``IOExists`` for defining 6 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1, T2, T3, T4, T5, T6], Any]) -> bool:
        pass


class IOExists7(Generic[T1, T2, T3, T4, T5, T6, T7]):
    """``IOExists`` for defining 7 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1, T2, T3, T4, T5, T6, T7], Any]) -> bool:
        pass


class IOExists8(Generic[T1, T2, T3, T4, T5, T6, T7, T8]):
    """``IOExists`` for defining 8 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7],
            t8: Type[T8]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1, T2, T3, T4, T5, T6, T7, T8], Any]) -> bool:
        pass


class IOExists9(Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9]):
    """``IOExists`` for defining 9 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7],
            t8: Type[T8],
            t9: Type[T9]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9], Any]) -> bool:
        pass


class IOExists10(Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]):
    """``IOExists`` for defining 10 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7],
            t8: Type[T8],
            t9: Type[T9],
            t10: Type[T10]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[
                [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10],
                Any]) -> bool:
        pass


class IOExists11(Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11]):
    """``IOExists`` for defining 11 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7],
            t8: Type[T8],
            t9: Type[T9],
            t10: Type[T10],
            t11: Type[T11]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[
                [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11],
                Any]) -> bool:
        pass


class IOExists12(Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12]):
    """``IOExists`` for defining 12 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7],
            t8: Type[T8],
            t9: Type[T9],
            t10: Type[T10],
            t11: Type[T11],
            t12: Type[T12]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[
                [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12],
                Any]) -> bool:
        pass


class IOExists13(Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12,
                         T13]):
    """``IOExists`` for defining 13 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7],
            t8: Type[T8],
            t9: Type[T9],
            t10: Type[T10],
            t11: Type[T11],
            t12: Type[T12],
            t13: Type[T13]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[
                [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13],
                Any]) -> bool:
        pass


class IOExists14(Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12,
                         T13, T14]):
    """``IOExists`` for defining 14 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7],
            t8: Type[T8],
            t9: Type[T9],
            t10: Type[T10],
            t11: Type[T11],
            t12: Type[T12],
            t13: Type[T13],
            t14: Type[T14]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[
                [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13,
                 T14],
                Any]) -> bool:
        pass


class IOExists15(Generic[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12,
                         T13, T14, T15]):
    """``IOExists`` for defining 15 IO existential variables."""

    def __init__(
            self,
            t1: Type[T1],
            t2: Type[T2],
            t3: Type[T3],
            t4: Type[T4],
            t5: Type[T5],
            t6: Type[T6],
            t7: Type[T7],
            t8: Type[T8],
            t9: Type[T9],
            t10: Type[T10],
            t11: Type[T11],
            t12: Type[T12],
            t13: Type[T13],
            t14: Type[T14],
            t15: Type[T15]) -> None:
        pass

    def __call__(
            self,
            expr: Callable[
                [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13,
                 T14, T15],
                Any]) -> bool:
        pass


def Open(operation: bool) -> None:
    """Replace non basic IO operation with its contents."""


def GetGhostOutput(operation: bool, output_name: str) -> Any:
    """Bind an expected IO operation output to a variable.

    Example usage::

        t2 = GetGhostOutput(read_int_io(t1), 't_post')
    """


def IOOperation(func: T) -> T:
    """Decorator to mark IO operations. It's a no-op."""
    return func


def Terminates(value: bool) -> None:
    """Provide IO operation termination condition."""


def TerminationMeasure(value: int) -> None:
    """Provide IO operation termination measure."""


class Place:
    """Denotes a place in a Petri Net."""


TOP_MEASURE = -1


def token(t: Place, measure: int = TOP_MEASURE) -> bool:
    """Indicate that there is a token at place ``t``."""


def ctoken(t: Place) -> bool:
    """Indicate that there is a credit token at place ``t``."""


__all__ = (
    'IOExists1',
    'IOExists2',
    'IOExists3',
    'IOExists4',
    'IOExists5',
    'IOExists6',
    'IOExists7',
    'IOExists8',
    'IOExists9',
    'IOExists10',
    'IOExists11',
    'IOExists12',
    'IOExists13',
    'IOExists14',
    'IOExists15',
    'Open',
    'GetGhostOutput',
    'IOOperation',
    'Terminates',
    'TerminationMeasure',
    'Place',
    'token',
    'ctoken',
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Public interface to IO operation translator."""


import ast

from typing import List, Tuple

from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonIOOperation,
)
from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.typedefs import (
    Function,
    Method,
    Predicate,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.abstract import TranslatorConfig
from nagini_translation.translators.io_operation.definition import (
    IOOperationDefinitionTranslator,
)
from nagini_translation.translators.io_operation.use import (
    IOOperationUseTranslator,
)


class IOOperationTranslator:
    """Class providing interface to translating IO operations."""

    def __init__(self, config: TranslatorConfig, jvm: JVM, source_file: str,
                 type_info: TypeInfo, viper_ast: ViperAST) -> None:
        self._use_translator = IOOperationUseTranslator(
            config, jvm, source_file, type_info, viper_ast)
        self._definition_translator = IOOperationDefinitionTranslator(
            config, jvm, source_file, type_info, viper_ast)

    def translate_io_operation(
            self, operation: PythonIOOperation,
            ctx: Context) -> Tuple[
                Predicate,
                List[Function],
                List[Method]]:
        """Translate IO operation to Silver."""
        return self._definition_translator.translate_io_operation(
            operation, ctx)

    def translate_io_operation_call(self, node: ast.Call,
                                    ctx: Context) -> StmtsAndExpr:
        """Translate a call to an IO operation."""
        return self._use_translator.translate_io_operation_call(
            node, ctx)

    def is_io_existential_defining_equality(self, node: ast.expr,
                                            ctx: Context) -> bool:
        """Check if ``node`` defines IO existential variable."""
        return self._use_translator.is_io_existential_defining_equality(
            node, ctx)

    def define_io_existential(self, node: ast.Compare, ctx: Context) -> None:
        """From defining equality defines IO existential variable."""
        self._use_translator.define_io_existential(
            node, ctx)

    def translate_get_ghost_output(self, node: ast.Assign,
                                   ctx: Context) -> List[Stmt]:
        """Translate ``GetGhostOutput``."""
        return self._use_translator.translate_get_ghost_output(
            node, ctx)

    def translate_io_contractfunc_call(self, node: ast.Call,
                                       ctx: Context) -> StmtsAndExpr:
        """Translate a call to a IO contract function."""
        return self._use_translator.translate_io_contractfunc_call(
            node, ctx)
